/*
 * SD_IDE64_BAM.c
 *
 * holds the BAM related functions
 */


#include "SD_IDE64.h"

/*
 * SD_IDE64_BAMByte()
 *
 * RETURNS: the byte, which holds the BAM bit
 * for the given logical block. This version includes
 * also a "caching for poor people" mechanism...
 */
int SD_IDE64_BAMByte(unsigned long int logicalblock, int action, unsigned char bambyte)
{
	int debug = 0;
	static long int lastaccessedbamblocknumber = -1;
	static unsigned char lastaccessedbamblock[BLOCKSIZE];
	unsigned long int bambyteoffset, bambytelogicalblocknumber;
	struct chs tempchs;

	if(action == BAM_FLUSH)
	{
		if(flags.verbose)
			fprintf(stdout, "Trying to flush BAM block cache... ");
		if(lastaccessedbamblocknumber != -1)
		{
			if(SD_IDE64_BlockWrite(phddinfo->pdevicefh, SD_IDE64_LogicalBlock2CHS(lastaccessedbamblocknumber, &tempchs), lastaccessedbamblock))
			{
				if(flags.verbose)
					printf("OK.\n");
				return 0;
			}
			else
			{
				if(flags.verbose)
					printf("ERROR!\n");
				fprintf(stderr, "SD_IDE64_BAMByte() - BAM_FLUSH: error writing block back.\n");
				return -1;
			}
		}
		else
		{
			if(flags.verbose)
				printf("OK.\n");
			if(flags.debug && debug)
				fprintf(stdout, "(No block to flush)\n");
			return 0;
		}
	}

	// OK, we are here to do something more than flushing the cache only
	bambytelogicalblocknumber = SD_IDE64_CHS2LogicalBlock(&phddinfo->bam1start) + (logicalblock / BAMBITSPERBYTE / phddinfo->blocksize);
	bambyteoffset = (logicalblock / BAMBITSPERBYTE) % phddinfo->blocksize;

	if(debug)
		fprintf(stdout, "SD_IDE64_BAMByte():\t bambase = %ld, bambyteblock = %ld, offset = %ld\n", SD_IDE64_CHS2LogicalBlock(&phddinfo->bam1start), bambytelogicalblocknumber, bambyteoffset);

	if(bambytelogicalblocknumber != lastaccessedbamblocknumber)
	{
		if(lastaccessedbamblocknumber != -1)
		{
			if(!flags.readonly)
			{
				if(!SD_IDE64_BlockWrite(phddinfo->pdevicefh, SD_IDE64_LogicalBlock2CHS(lastaccessedbamblocknumber, &tempchs), lastaccessedbamblock))
				{
					fprintf(stderr, "SD_IDE64_BAMByte() - BAM_WRITE: error writing block back.\n");
					return -1;
				}
			}
		}
		if(SD_IDE64_BlockRead(phddinfo->pdevicefh, SD_IDE64_LogicalBlock2CHS(bambytelogicalblocknumber, &tempchs), lastaccessedbamblock))
		{
			lastaccessedbamblocknumber = bambytelogicalblocknumber;
		}
		else
		{
			fprintf(stderr, "SD_IDE64_BAMByte() - Error reading BAM block.\n");
			return -1;
		}
	}
	if(action == BAM_WRITE)
		lastaccessedbamblock[bambyteoffset] = bambyte;
	return lastaccessedbamblock[bambyteoffset];
}



/*
 * SD_IDE64_GetBAMBit()
 *
 * RETURNS: the state of the BAM bit for the given CHS block
 */
int SD_IDE64_GetBAMBit(struct chs *pchs)
{
	int result, bambyte, bitnumber;
	bambyte = SD_IDE64_BAMByte(SD_IDE64_CHS2LogicalBlock(pchs), BAM_READ, 0);
	if(bambyte < 0)
	{
		fprintf(stderr, "SD_IDE64_GetBAMBit(): can't get BAMbit for %d/%d/%d\n", pchs->cylhi * 256 + pchs->cyllo, pchs->head, pchs->sec);
		result = -1;
	}
	else
	{
		bitnumber = SD_IDE64_CHS2LogicalBlock(pchs) % 8;
		result = (bambyte & (1 << bitnumber)) >> bitnumber;
	}
	return result;
}



/*
 * SD_IDE64_SetBAMBit()
 *
 * Sets or clears the BAM bit for the given CHS block
 * RETURNS: the resulting value of the bit or -1 on error
 */
int SD_IDE64_SetBAMBit(struct chs *pchs, int setbitto)
{
	int result, bambyte, bitnumber;
	bambyte = SD_IDE64_BAMByte(SD_IDE64_CHS2LogicalBlock(pchs), BAM_READ, 0);
	if(bambyte < 0)
	{
		fprintf(stderr, "SD_IDE64_SetBAMBit(): can't get BAMbit for %d/%d/%d\n", pchs->cylhi * 256 + pchs->cyllo, pchs->head, pchs->sec);
		result = -1;
	}
	else
	{
		bitnumber = SD_IDE64_CHS2LogicalBlock(pchs) % 8;

		if(setbitto)
			bambyte = bambyte | (1 << bitnumber);
		else
			bambyte = bambyte & ((1 << bitnumber) ^ 0xff);

		if(SD_IDE64_BAMByte(SD_IDE64_CHS2LogicalBlock(pchs), BAM_WRITE, bambyte) >= 0)
			result = SD_IDE64_GetBAMBit(pchs);
		else
			result = -1;
	}
	return result;
}



/*
 * SD_IDE64_FindFreeBlock()
 *
 * RETURNS: the number of the first available logical
 * block number, 0 if DISK FULL, -1 on error.
 */
long int SD_IDE64_FindFreeBlock(void)
{
	long int i, startinglogicalblock;
	struct chs tempchs, *ptempchs = &tempchs;

	startinglogicalblock = SD_IDE64_CHS2LogicalBlock(&phddinfo->root_direntry.filestartchs);
	for(i = 0; i < (phddinfo->numberoflogicalblocks - startinglogicalblock); i++)
	{
		if(SD_IDE64_LogicalBlock2CHS(startinglogicalblock + i, ptempchs))
		{
			if(!SD_IDE64_GetBAMBit(ptempchs))
				return startinglogicalblock + i;
		}
		else
			return -1;
	}
	return 0;
}




/*
 * SD_IDE64_FindFreeBlocks()
 *
 * RETURNS: the first available logical block number
 * of a continuous sequence of free blocks at least
 * numofblockstofind long.
 * 0 if DISK FULL, -1 on error.
 */
long int SD_IDE64_FindFreeBlocks(int numofblockstofind)
{
	long int i, j, startinglogicalblock;
	struct chs tempchs, *ptempchs = &tempchs;

	startinglogicalblock = SD_IDE64_CHS2LogicalBlock(&phddinfo->root_direntry.filestartchs);
	for(i = 0; i < (phddinfo->numberoflogicalblocks - startinglogicalblock); i++)
	{
		if(SD_IDE64_LogicalBlock2CHS(startinglogicalblock + i, ptempchs))
		{
			if(!SD_IDE64_GetBAMBit(ptempchs))
			{
				for(j = 0; j < numofblockstofind; j++)
				{
					if(SD_IDE64_LogicalBlock2CHS(startinglogicalblock + i +j, ptempchs))
					{
						if(SD_IDE64_GetBAMBit(ptempchs) > 0)
							break;
					}
				}
				if(j == numofblockstofind)
					return startinglogicalblock + i;
			}
		}
		else
			return -1;
	}
	return 0;
}





/*
 * SD_IDE64_BlocksFree()
 *
 * Calculates the number of HDD blocks marked as FREE
 * RETURNS: long integer number of blocks free.
 */
long int SD_IDE64_BlocksFree(int format)
{
	long int i, result, startinglogicalblock;
	struct chs tempchs, *ptempchs = &tempchs;

	result = 0;
	startinglogicalblock = SD_IDE64_CHS2LogicalBlock(&phddinfo->root_direntry.filestartchs);
	for(i = 0; i < (phddinfo->numberoflogicalblocks - startinglogicalblock); i++)
	{
		if(SD_IDE64_LogicalBlock2CHS(startinglogicalblock + i, ptempchs))
		{
			if(!SD_IDE64_GetBAMBit(ptempchs))
				result++;
		}
		else
			result = -1;
	}
	if(format)
	{
		result = SD_IDE64_Bytes2CBMBlocks(phddinfo->blocksize * result);
	}
	return result;
}
