/*
 * SD_IDE64_Commands.c
 *
 * holds all the command line commands' routines
 */

#include "SD_IDE64.h"






int DoTest(int argc, char *argv[]) {
	int i;
	printf ("This is a placeholder for test and debug routines.\nShowing arguments:\n");
	i=0;
	while (i<argc) {
		printf("%i:%s:\n",i,argv[i]);
		i++;
	}
	printf("\n");
	return EXIT_SUCCESS;
}






int DoLs(int argc, char *argv[]) {

	int i = 0, entries = 0, result = EXIT_SUCCESS;
	char *ppathbuffer;

	do
	{
		if(argc)
			ppathbuffer = malloc((strlen(argv[i]) + 1));
		else
		{
			if(flags.cnvexts)
				ppathbuffer = malloc((strlen("*.*") + 1));
			else
				ppathbuffer = malloc((strlen("*") + 1));
		}
		if(ppathbuffer)
		{
			if(argc)
				strcpy(ppathbuffer, argv[i]);
			else
			{
				if(flags.cnvexts)
					strcpy(ppathbuffer, "*.*");
				else
					strcpy(ppathbuffer, "*");
			}

			entries += SD_IDE64_FSNewRTraverse(ppathbuffer, &phddinfo->cur_direntry.filestartchs, TRAVERSE_DISPLAY);

			if(entries < 0)
			{
				fprintf(stderr, "ls - error scanning directory: %s\n", ppathbuffer);
				entries = 0;
				result = EXIT_FAILURE;
			}
			if(!flags.recursive)
				fprintf(stdout, "Entries listed:\t%d\n", entries);

			free(ppathbuffer);
		}
		else
		{
			fprintf(stderr, "ls - ");
			perror("malloc");
			result = EXIT_FAILURE;
			break;
		}
		SD_IDE64_DirectoryCD2Root();
		i++;
	} while(i < argc);

	return result;
}




int DoMkdir(int argc, char *argv[]) {
	int i;

	if(flags.readonly)
	{
		fprintf(stderr, "mkdir - error: read only filesystem!\n");
		return EXIT_FAILURE;
	}

    i=0;
    while (i<argc) {
		if(SD_IDE64_DirectoryMk(argv[i]))
			return EXIT_FAILURE;
		i++;
    }
    return EXIT_SUCCESS;
}





int DoBackup(int argc, char *argv[]) {

	/* STATUS: EXPERIMENTAL! */
    int i;
    char filenamebuffer[256];
    struct stat sb;

	for(i = 0; i < 64; i++) //we don't want to create zillions of dirs when the bugs hit us badly...
	{
		if ((i==0)&&(outputname)) {
			strcpy(filenamebuffer, outputname);	/* first try supplied name */
		} else {
			sprintf(filenamebuffer, "ide64backup_%03d", i);
		}
		if(flags.verbose)
			printf("Checking %s... ", filenamebuffer);
		if(!stat(filenamebuffer, &sb))
		{
			if(flags.verbose)
				printf("EXISTS!\n");
		}
		else
		{
			if(errno == ENOENT)
			{
				if(flags.verbose)
					printf("OK.\n");
			}
			if(flags.verbose)
				printf("Creating directory %s... ", filenamebuffer);
			if(mkdir(filenamebuffer, 0755))
			{
				if(flags.verbose)
					printf("BAD LUCK!\n");
				fprintf(stderr, "main() - ");
				perror("mkdir");
				flags.error = 1;
			}
			else
			{
				if(flags.verbose)
					printf("OK.\nGoing there...");
				if(chdir(filenamebuffer))
				{
					if(flags.verbose)
						printf("BAD LUCK\n");
					fprintf(stderr, "main() - ");
					perror("chdir");
					flags.error = 1;
				}
				else
				{
					if(flags.verbose)
						printf("OK.\n");
					break;
				}
			}
		}
	}

	if(!flags.error)
		fprintf(stdout, "Entries processed:\t%d\n", SD_IDE64_FSRTraverse(&phddinfo->cur_direntry.filestartchs, NULL, TRAVERSE_BACKUP));

	if(flags.error)	{
		return EXIT_FAILURE;
	} else {
		return EXIT_SUCCESS;
	}
}





int DoRestore(int argc, char *argv[]) {

	/* STATUS: EXPERIMENTAL! */
	int result = EXIT_SUCCESS;
	long int entriesrestored = 0;

	if(!flags.readonly)
	{
		entriesrestored = SD_IDE64_LocalRTraverse(argv[0], TRAVERSE_RESTORE);
		if(entriesrestored < 0)
		{
			fprintf(stderr, "restore - error during restore.\n");
			result = EXIT_FAILURE;
		}
	}
	else
	{
		fprintf(stderr, "restore - error: read only filesystem!\n");
		result = EXIT_FAILURE;
	}

	printf("Entries restored: %ld\n", entriesrestored);
	return result;
}





int DoFsck(int argc, char *argv[]) {
	fprintf(stderr, "Not implemented yet... \n");
    return EXIT_FAILURE;
}





int DoMkfs(int argc, char *argv[]) {
	fprintf(stderr, "Not implemented yet... \n");
    return EXIT_FAILURE;
}





int DoMv(int argc, char *argv[]) {
	fprintf(stderr, "Not implemented yet... \n");
    return EXIT_FAILURE;
}





int DoInfo(int argc, char *argv[]) {
	long int freehddblocks;
	float megabytes;
	SD_IDE64_PrintDriveInfo(phddinfo);
	printf("Counting blocks (this may take a while)... \n");
	freehddblocks = SD_IDE64_BlocksFree(HDDBLOCKS);
	megabytes = (phddinfo->numberoflogicalblocks - freehddblocks) / 2000; // IMPROVEME: should we take the filesystem overhead into account?
	printf("%.2f MiB / ", megabytes);
	printf("%ld HDD / ", (phddinfo->numberoflogicalblocks - freehddblocks));
	printf("%ld CBM BLOCKS USED\n", SD_IDE64_Bytes2CBMBlocks(phddinfo->blocksize * (phddinfo->numberoflogicalblocks - freehddblocks)));
	megabytes = freehddblocks / 2000; // IMPROVEME: should we take the filesystem overhead into account?
	printf("%.2f MiB / ", megabytes);
	printf("%ld HDD / ", freehddblocks);
	printf("%ld CBM BLOCKS FREE\n", SD_IDE64_Bytes2CBMBlocks(phddinfo->blocksize * freehddblocks));
	return EXIT_SUCCESS;
}





int DoRm(int argc, char *argv[]) {

	// STATUS: EXPERIMENTAL!

	// Be very cautious and keep
	// your backup always at hand!

	struct ide64node ide64node;
	unsigned char filename[FILENAMELENGTH +1];
	char *pnamepointer, *pnamebuffer;
	int i = 0, entries = 0, entriesfound, result = EXIT_SUCCESS;

	if(flags.readonly)
	{
		fprintf(stderr, "rm - error: read only filesystem!\n");
		return EXIT_FAILURE;
	}

    while (i<argc)
	{
		pnamepointer = SD_IDE64_Filename(argv[i]);
		if(pnamepointer)
		{
			if(strlen(pnamepointer))
			{
				pnamebuffer = malloc(strlen(pnamepointer) + 1);
				if(pnamebuffer)
				{
					strcpy(pnamebuffer, pnamepointer);
					memcpy(&ide64node.dirblockchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs));
					ide64node.direntryindex = 0;
					ide64node.direntry.attribute = SD_IDE64_Filename2FileAttByte(pnamepointer);

					if(flags.cnvexts && ide64node.direntry.attribute)
						stripExtension(pnamepointer);
					else
						ide64node.direntry.attribute = FILEATT_ANY;

					memcpy(&ide64node.direntry.filename[0], SD_IDE64_Filename2CBMFilename(filename, pnamepointer), FILENAMELENGTH);

					while((entriesfound = SD_IDE64_DirectoryFindEntries(&ide64node)) > 0)
					{
						if(!(ide64node.direntry.attribute & FILEATT_DIR))
						{
							if(!(SD_IDE64_NodeDelete(&ide64node)))
							{
								entries++;
								ide64node.direntryindex++;
								strcpy(pnamepointer, pnamebuffer);
								ide64node.direntry.attribute = SD_IDE64_Filename2FileAttByte(pnamepointer);

								if(flags.cnvexts && ide64node.direntry.attribute)
									stripExtension(pnamepointer);
								else
									ide64node.direntry.attribute = FILEATT_ANY;

								memcpy(&ide64node.direntry.filename[0], SD_IDE64_Filename2CBMFilename(filename, pnamepointer), FILENAMELENGTH);
							}
						}
						else
						{
							fprintf(stderr, "rm - %s: is a directory\n", pnamepointer);
							result = EXIT_FAILURE;
							break;
						}
					}
					if(entriesfound < 0)
					{
						fprintf(stderr, "rm - error seearching directory.\n");
						result = EXIT_FAILURE;
						break;
					}
					free(pnamebuffer);
				}
				else
				{
					fprintf(stderr, "rm - ");
					perror("malloc");
					result = EXIT_FAILURE;
					break;
				}
			}
			else
			{
				fprintf(stderr, "rm - filename [pattern] required\n");
				result = EXIT_FAILURE;
				break;
			}
		}
		else
		{
			fprintf(stderr, "rm - valid path/filename required.\n");
			result = EXIT_FAILURE;
			break;
		}
		if(i == (argc - 1))
			fprintf(stdout, "FILES SCRATCHED:\t%d\n", entries);
		i++;
    }
	return result;
}





int DoRmDir(int argc, char *argv[]) {

	// STATUS: EXPERIMENTAL!

	// Be very cautious and keep
	// your backup always at hand!

	struct ide64node ide64node;
	unsigned char filename[FILENAMELENGTH +1];
	char *pnamepointer;
	int i, entries = 0, entriesfound, entriesindir, result = EXIT_SUCCESS;

	if(flags.readonly)
	{
		fprintf(stderr, "rmdir - error: read only filesystem!\n");
		return EXIT_FAILURE;
	}

    i=0;
    while (i<argc) {
		pnamepointer = SD_IDE64_Filename(argv[0]);
		if(pnamepointer)
		{
			if(strlen(pnamepointer) == 0)
			{
				fprintf(stderr, "rmdir - filename [pattern] required\n");
				result = EXIT_FAILURE;
				break;
			}
			memcpy(&ide64node.dirblockchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs));
			ide64node.direntryindex = 0;
			ide64node.direntry.attribute = SD_IDE64_Filename2FileAttByte(pnamepointer);

			if(flags.cnvexts && ide64node.direntry.attribute)
				stripExtension(pnamepointer);

			ide64node.direntry.attribute = FILEATT_DIR;			

			memcpy(&ide64node.direntry.filename[0], SD_IDE64_Filename2CBMFilename(filename, pnamepointer), FILENAMELENGTH);

			while((entriesfound = SD_IDE64_DirectoryFindEntries(&ide64node)) > 0)
			{
				if((ide64node.direntry.attribute & FILEATT_DIR))
				{
					entriesindir = SD_IDE64_DirectoryScan(&ide64node.direntry.filestartchs, 0);
					if(entriesindir == 2)
					{
						if(flags.verbose)
							printf("Deleting %s...\n", pnamepointer);
						if(!(SD_IDE64_NodeDelete(&ide64node)))
						{
							entries++;
							ide64node.direntryindex++;
							memcpy(&ide64node.direntry.filename[0], SD_IDE64_Filename2CBMFilename(filename, pnamepointer), FILENAMELENGTH);

							ide64node.direntry.attribute = FILEATT_DIR;
						}

					}
					else
					{
						fprintf(stderr, "rmdir - %s: Directory not empty\n", pnamepointer);
						result = EXIT_FAILURE;
						break;
					}
				}
				else
				{
					fprintf(stderr, "rmdir - %s: Not a directory\n", pnamepointer);
					result = EXIT_FAILURE;
					break;
				}
			}
			if(entriesfound < 0)
			{
				fprintf(stderr, "rm - error seearching directory.\n");
				result = EXIT_FAILURE;
			}
		}
		else
		{
			fprintf(stderr, "rm - valid path/filename required.\n");
			result = EXIT_FAILURE;
		}
		if(i == (argc - 1))
			fprintf(stdout, "DIRS SCRATCHED:\t%d\n", entries);
		i++;
    }
	return result;
}





int DoRead(int argc, char *argv[])
{

	// STATUS: EXPERIMENTAL!

	struct ide64node ide64node;
	FILE *pfile;
	unsigned char filename[FILENAMELENGTH + 1 + 3 + 1 + 1]; // FILENAME + "." + EXTENSION + LOCKED? + \0
	char *pnamepointer;
	int entries = 0, found, result = EXIT_SUCCESS;

	if(argc)
	{
		pnamepointer = SD_IDE64_Filename(argv[0]);
		if(pnamepointer)
		{
			if(strlen(pnamepointer) > 0)
			{
				memcpy(&ide64node.dirblockchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs));
				ide64node.direntryindex = 0;
				ide64node.direntry.attribute = SD_IDE64_Filename2FileAttByte(pnamepointer);

				if(flags.cnvexts && ide64node.direntry.attribute)
					stripExtension(pnamepointer);
				else
					ide64node.direntry.attribute = FILEATT_ANY;

				memcpy(&ide64node.direntry.filename[0], SD_IDE64_Filename2CBMFilename(filename, pnamepointer), FILENAMELENGTH);

				found = SD_IDE64_DirectoryFindEntries(&ide64node);
				if(found > 0)
				{
					if(!(ide64node.direntry.attribute & FILEATT_DIR))
					{
						if(ide64node.direntry.attribute & FILEATT_CLOSED)
						{
							if(argc > 1)
								pfile = fopen(argv[1], "w");
							else
								pfile = fopen(SD_IDE64_CBMFilename2Filename(filename, ide64node.direntry.filename, ide64node.direntry.attribute), "w");

							if(flags.verbose)
								printf("%s... ", filename);

							if(pfile)
							{
								if(SD_IDE64_FileRead(pfile, &ide64node))
								{
									fprintf(stderr, "read - error reading file.\n");
									result = EXIT_FAILURE;
								}

								if(!(fclose(pfile)))
								{
									entries++;
									if(flags.verbose)
										printf("OK\n");
								}
								else
								{
									fprintf(stderr, "read - error closing the destination file.\n");
									perror("fclose");
									result = EXIT_FAILURE;
								}
							}
							else
							{
								if(argc > 1)
									fprintf(stderr, "read - error opening destination: %s\n", argv[1]);
								else
									fprintf(stderr, "read - error opening destination: %s\n", filename);
								perror("fopen");
								result = EXIT_FAILURE;
							}
						}
						else
						{
							fprintf(stderr, "read - FILE OPEN.\n");
							result = EXIT_FAILURE;
						}
					}
					else
					{
						fprintf(stderr, "read - %s is a directory!\n", pnamepointer);
						result = EXIT_FAILURE;
					}
				}
				else
				{
					fprintf(stderr, "read - error searching for: %s\n", argv[0]);
					result = EXIT_FAILURE;
				}
			}
			else
			{
				fprintf(stderr, "read - valid filename required.\n");
				result = EXIT_FAILURE;
			}
		}
		else
		{
			fprintf(stderr, "read - valid path/filename required.\n");
			result = EXIT_FAILURE;
		}
	}
	else
	{
		fprintf(stderr, "read - path/filename required.\n");
		result = EXIT_FAILURE;
	}
	if(flags.verbose)
		fprintf(stdout, "FILES READ:\t%d\n", entries);
	return result;
}





int DoWrite(int argc, char *argv[])
{
	// STATUS: EXPERIMENTAL!

	int debug = 1;
	FILE *pfile;
	IDE64FILE *pide64file;
	unsigned char attribute = 0;
	char *pnamepointer;
	int i, result = EXIT_SUCCESS;

	struct stat sb;

	if(flags.readonly)
	{
		fprintf(stderr, "write - error: read only filesystem!\n");
		return EXIT_FAILURE;
	}

	if(argc)
	{
		if(!(stat(argv[0], &sb)))	// FIXME: Is this needed?! To check the existence won't the proper fopen() be enough?!
		{
			pfile = fopen(argv[0], "r");
			if(pfile)
			{
				if(argc > 1)
					pnamepointer = SD_IDE64_Filename(argv[1]);
				else
					pnamepointer = filenameOnly(argv[0]);

				if(pnamepointer)
				{
					if(strlen(pnamepointer))
						attribute = SD_IDE64_Filename2FileAttByte(pnamepointer);
					else
						pnamepointer = filenameOnly(argv[0]);

					if(flags.cnvexts && attribute)
						stripExtension(pnamepointer);
					else
						attribute = FILEATT_PRG;

					pide64file = SD_IDE64_FileOpen(pnamepointer, attribute, "w");
					if(pide64file)
					{
						// FIXME: check for errors on all those!
						i = SD_IDE64_FileWrite(pide64file, pfile);
						if(debug)
							printf("%d bytes written.\n", i);
						SD_IDE64_FileClose(pide64file);
					}
				}
				else
				{
					fprintf(stderr, "write - valid path/filename expected.\n");
					result = EXIT_FAILURE;
				}

				if(fclose(pfile))
				{
					fprintf(stderr, "write - ");
					perror("fclose");
					result = EXIT_FAILURE;
				}
			}
			else
			{
				fprintf(stderr, "write - ");
				perror("fopen");
				result = EXIT_FAILURE;
			}
		}
		else
		{
			fprintf(stderr, "write - ");
			perror("stat");
			result = EXIT_FAILURE;
		}
	}
	else
	{
		fprintf(stderr, "write - path/filename required.\n");
		result = EXIT_FAILURE;
	}

	return result;
}





