/*
 * SD_IDE64_Directory.c
 *
 * Holds the functions related to the drive's DIRECTORY
 */


#include "SD_IDE64.h"


/*
 * SD_IDE64_Bytes2CBMBlocks()
 *
 * calculates the number of CBM DOS 2.6 blocks
 * required for storing the given number of bytes
 * RETURNS: long integer number of CBM DOS 2.6 blocks
 */
long int SD_IDE64_Bytes2CBMBlocks(long int numofbytes)
{
	int numofcbmblocks;
	numofcbmblocks = numofbytes%254 ? (numofbytes/254)+1 : numofbytes/254;
	return numofcbmblocks;
}

/*
 * SD_IDE64_Bytes2HDDBlocks()
 * calculates the number of harddrive's blocks required
 * for storing the given number of bytes
 */
long int SD_IDE64_Bytes2HDDBlocks(long int numofbytes)
{
	int numofhddblocks;
	numofhddblocks = numofbytes % (phddinfo->blocksize - sizeof(struct chs)) ? (numofbytes / (phddinfo->blocksize - sizeof(struct chs)))+1 : numofbytes / (phddinfo->blocksize - sizeof(struct chs));
	return numofhddblocks;
}





/*
 * SD_IDE64_DirectoryFindEntries()
 *
 * Searches for the direntries matching
 * the given (struct ide64node) citeria.
 * RETURNS: number of entries found. -1 on errors.
 */
int SD_IDE64_DirectoryFindEntries(struct ide64node *pide64node)
{
	int i, j, k, found, foundindex = 0, numofentries = 0;
	struct dirblock *pdirblock;
	struct direntry founddirentry;
	struct chs tempchs, foundchs;
	if((pdirblock = malloc(sizeof(struct dirblock))))
	{
		memcpy(&tempchs, &pide64node->dirblockchs, sizeof(struct chs));
		k = pide64node->direntryindex;
		do
		{
			if(SD_IDE64_BlockRead(phddinfo->pdevicefh, &tempchs, (unsigned char *)pdirblock))
			{
				for(i = k; i < NUMOFDIRENTRIES; i++)
				{
					if(pdirblock->direntry[i].attribute) // skip the empty or deleted entries
					{
						found = 0;
						if(((pide64node->direntry.attribute & (0xff ^ (FILEATT_CLOSED | FILEATT_LOCKED | FILEATT_LABEL)))
							== (pdirblock->direntry[i].attribute & (0xff ^ (FILEATT_CLOSED | FILEATT_LOCKED | FILEATT_LABEL))))
							|| (pide64node->direntry.attribute == FILEATT_ANY))
						{
							for(j = 0; j < FILENAMELENGTH; j++)
							{
								if(pide64node->direntry.filename[j] == '*') // IMPROVEME: should be able to use * anywhere.
								{											// IDE64, however, doesn't seem to care about it.
									found = 1;
									break;
								}
								if((pide64node->direntry.filename[j] == 0xa0) && (pdirblock->direntry[i].filename[j] == 0xa0))
								{
									found = 1;
									break;
								}
								if(pide64node->direntry.filename[j] != pdirblock->direntry[i].filename[j])
								{
									if((pide64node->direntry.filename[j] == '?') && (pdirblock->direntry[i].filename[j] != 0xa0))
									{
										if(j == (FILENAMELENGTH -1))
										{
											found = 1;
											break;
										}
										else
										{
											continue;
										}
									}
									else
										break;
								}
								if(j == (FILENAMELENGTH -1))
									found = 1;
							}
							if(found)
							{
								numofentries++;
								if(numofentries == 1)
								{
									memcpy(&founddirentry, &pdirblock->direntry[i], sizeof(struct direntry));
									memcpy(&foundchs, &tempchs, sizeof(struct chs));
									foundindex = i;
								}
							}
						}
					}
				}
				tempchs.cylhi = pdirblock->nextfileblock.cylhi & CLEAR2MSBITSOF8;
				tempchs.cyllo = pdirblock->nextfileblock.cyllo;
				tempchs.head = pdirblock->nextfileblock.head;
				tempchs.sec = pdirblock->nextfileblock.sec;
				k = 0;
			}
			else
			{
				fprintf(stderr, "SD_IDE64_DirectoryFindEntries() - Error reading block [%d/%d/%d]\n", tempchs.cylhi * 256 + tempchs.cyllo, tempchs.head, tempchs.sec);
				numofentries = -1;
				free(pdirblock);
				break;
			}
		} while(!(pdirblock->nextfileblock.cylhi & BIT6OF8)); // isolating bit 6, checking for more blocks
		free(pdirblock);
	}
	else
	{
		fprintf(stderr, "SD_IDE64_DirectoryFindEntries() - ");
		perror("malloc");
		numofentries = -1;
	}
	if(numofentries)
	{
		memcpy(&pide64node->direntry, &founddirentry, sizeof(struct direntry));
		memcpy(&pide64node->dirblockchs, &foundchs, sizeof(struct chs));
		pide64node->direntryindex = foundindex;
	}
	return numofentries;
}





/*
 * SD_IDE64_DirectoryFindIn()
 *
 * scans the current directory in search of the given filename AND fileatt
 * modifies the pfoundchs pointer to point to the appropriate dirblock if needed.
 * RETURNS: the direntry index within the pfoundchs dirblock or -1 on error (or NOT FOUND)
 */
int SD_IDE64_DirectoryFindIn(char *pfilename, unsigned char fileatt, struct chs *pfoundchs)
{
	int debug = 0;
	struct dirblock *pdirblock;
	struct chs tempchs, *ptempchs = &tempchs;
	int i, j;
	int numberofdirblocks = 1, moreblocks = 0;
	unsigned char FilenameBuffer[21];

	memset(FilenameBuffer, 0xa0, sizeof(FilenameBuffer));
	for(i = 0; i < strlen(pfilename); i++)
		FilenameBuffer[i] = *(pfilename + i);

	memcpy(ptempchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs));

	pdirblock = malloc(sizeof(struct dirblock));
	if(pdirblock)
	{
		do
		{
			if(debug)
				printf("Reading dirblock...\n");
			moreblocks = SD_IDE64_ReadFileBlock(ptempchs, (struct fileblock *)pdirblock);
			if(moreblocks >= 0)
			{
				// The first entry in the first block of the directory should
				// be a label. If it isn't, then it doesn't look too promising...
				if(numberofdirblocks == 1)
				{
					if(!((pdirblock->direntry[0].attribute & (FILEATT_LABEL | FILEATT_DIR | FILEATT_LOCKED | FILEATT_CLOSED)) == (FILEATT_LABEL | FILEATT_DIR | FILEATT_LOCKED | FILEATT_CLOSED)))
						fprintf(stderr, "Warning: No label entry - [%d/%d/%d] doesn't look like a valid starting directory block...\n",ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
				}

				for (i = 0; i < NUMOFDIRENTRIES; i++)
				{
					if(pdirblock->direntry[i].attribute)
					{
						for (j = 0; j < sizeof(pdirblock->direntry[i].filename); j++)
						{
							if (SD_IDE64_PETASCII2ASCII(pdirblock->direntry[i].filename[j]) != FilenameBuffer[j])
							{
								break;
							}
						}
						// FIXME: Something concerning the extensions has to be done here!
						if((j == sizeof(pdirblock->direntry[i].filename)) && (((pdirblock->direntry[i].attribute & fileatt) == fileatt) || fileatt == FILEATT_ANY))
						{
							memcpy(pfoundchs, ptempchs, sizeof(struct chs));
							return i;
						}
					}
				}
				numberofdirblocks++;

				ptempchs->cylhi = pdirblock->nextfileblock.cylhi & CLEAR2MSBITSOF8;
				ptempchs->cyllo = pdirblock->nextfileblock.cyllo;
				ptempchs->head = pdirblock->nextfileblock.head;
				ptempchs->sec = pdirblock->nextfileblock.sec;

			}
			else
			{
				fprintf(stderr, "Error: SD_IDE64_DirectoryFindIn() - Can't read the block [%d/%d/%d]\n", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
				errno = EIO;
				return -1;
			}
		} while(moreblocks > 0);
		free(pdirblock);
	}
	else
	{
		fprintf(stderr, "Error: SD_IDE64_DirectoryFindIn() - Can't allocate memory for the block buffer\n");
		perror("malloc");
		pfoundchs = NULL;
		errno = ENOMEM;
		return -1;
	}
	pfoundchs = NULL;
	errno = ENOENT;
	return -1;
}





/*
 * SD_IDE64_DisplayDirEntry()
 *
 * Displays the entry given by the pdirentry pointer
 * RETURNS: 1. FIXME: strange... always returns true? Should we change this to void?
 */
int SD_IDE64_DisplayDirEntry(struct direntry *pdirentry)
{
	char ExtensionString[6], *pExtensionString = &ExtensionString[0];
	int filelengthinbytes, filenamequoteopen;
	int j;
//	mode_t localst_mode = phddinfo->sb.st_mode & (0xffff ^ (S_IFREG | S_IFBLK));
	mode_t localst_mode = phddinfo->sb.st_mode & (-1 ^ (S_IFREG | S_IFBLK));


	filelengthinbytes = 0;
	for (j=0; j<3; j++)
		filelengthinbytes += pdirentry->filesize[j] * power(256, j);


	if(flags.longdir)
	{
		if(pdirentry->attribute & FILEATT_DIR)
			localst_mode |= (S_IFDIR | S_IXUSR | S_IXGRP | S_IXOTH);
		else
			localst_mode |= S_IFREG;

		if(pdirentry->attribute & FILEATT_LOCKED)					// FIXME: This will cut off the w bits on locked files
			localst_mode &= (-1 ^ (S_IWUSR | S_IWGRP | S_IWOTH));	// but this is inconsistent with unix ways...

		printf("%9s ", bb_mode_string(localst_mode));
		printf("%3d ", 1);
		printf("%-8s ", phddinfo->ppasswd->pw_name);
		printf("%-8s ", phddinfo->pgroup->gr_name);
		printf("%8d ", filelengthinbytes);

		fprintf(stdout, "%02x-%02x-", pdirentry->timestamp_month, pdirentry->timestamp_day);
		// This is sick! How come people never learn and still use the year values w/o century?!
		if(pdirentry->timestamp_year > 70)
			fprintf(stdout, "19");
		else
			fprintf(stdout, "20");
		fprintf(stdout, "%02x ", pdirentry->timestamp_year);
		fprintf(stdout, "%02x:%02x ", pdirentry->timestamp_hour, pdirentry->timestamp_min);

		for (j = 0; j < sizeof(pdirentry->filename); j++)
		{
			if (pdirentry->filename[j] != 0xa0)
				fputc(SD_IDE64_PETASCII2ASCII(pdirentry->filename[j]), stdout);
			else
				break;
		}
		if(flags.cnvexts)
			fprintf(stdout, "%s", SD_IDE64_FileAttByte2Extension(pdirentry->attribute, pExtensionString, EXT4DISPLAY));
	}

	else

	{
		fprintf(stdout, "%-5ld", SD_IDE64_Bytes2CBMBlocks(filelengthinbytes));
		filenamequoteopen = 1;
		fprintf(stdout, "\"");
		for (j = 0; j < sizeof(pdirentry->filename); j++)
		{
			if (pdirentry->filename[j] == 0xa0)
			{
				if(filenamequoteopen)
				{
					fprintf(stdout, "\"");
					filenamequoteopen = 0;
					// This  "continue;" statement below is here
					// to mimic the CBM DOS behaviour. Due to a bug, it
					// prematurely closes the quote if the filename
					// contains the SHFT-SPACE ($A0) character
					// and displays the rest of the filename
					// after the closing quote. To change to "IDE64 mode"
					// replace "continue;" with "j++; break;"
					continue;
					// j++; break;
				}
				else
				{
					fputc(' ', stdout);
				}
			}
			else
			{
				fputc(SD_IDE64_PETASCII2ASCII(pdirentry->filename[j]), stdout);
			}
		}
		// If we haven't closed the quote yet because
		// the filename was long enough to fill the whole
		// space provided in the structure, we have to do it now
		// otherwise we add one SPACE to compensate for the
		// quote char used before
		if(filenamequoteopen)
		{
			fprintf(stdout, "\"");
			filenamequoteopen = 0;
		}
		else
		{
			fputc(' ', stdout);
		}
		// Now we have to pad the filename accordingly.
		// We use the loop here because it works with both
		// CBM and IDE64 "modes"
		for (j = sizeof(pdirentry->filename) - j; j > 0; j--)
		{
			fputc(' ', stdout);
		}
		fprintf(stdout, "%s", SD_IDE64_FileAttByte2Extension(pdirentry->attribute, pExtensionString, EXT4DISPLAY));
		if(flags.verbose)
		{
			fprintf(stdout, "\t%ld HDD blocks at: ", SD_IDE64_Bytes2HDDBlocks(filelengthinbytes));
			fprintf(stdout, "\t[%04d/%02d/%02d]", pdirentry->filestartchs.cylhi * 256 + pdirentry->filestartchs.cyllo, pdirentry->filestartchs.head, pdirentry->filestartchs.sec);
			// This is sick! Howcome people never learn and still use the year values w/o century?!
			if(pdirentry->timestamp_year > 70)
				fprintf(stdout, "\t19");
			else
				fprintf(stdout, "\t20");
			fprintf(stdout, "%02x-%02x-%02x", pdirentry->timestamp_year, pdirentry->timestamp_month, pdirentry->timestamp_day);
			fprintf(stdout, "\t%02x:%02x:%02x", pdirentry->timestamp_hour, pdirentry->timestamp_min, pdirentry->timestamp_sec);

		}
	}
	fprintf(stdout, "\n");
	return 1;
}





/*
 * SD_IDE64_DirectoryAllocEntry()
 *
 * Looks for a free entry in the current directory.
 * Modifies the given chs structure to point to the appropriate dirblock if needed
 * RETURNS: the index of the first free entry in the dirblock pointed to by ptempchs
 * or -1 on error.
 */
int SD_IDE64_DirectoryAllocEntry(struct chs *ptempchs)
{
	int i, numofentries = 0, freeentryindex = -1, moreblocks, numberofdirblocks = 1;
	struct dirblock *pdirblock;
	IDE64FILE ide64file;

	memcpy(ptempchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs));

	pdirblock = malloc(sizeof(struct dirblock));

	if(pdirblock)
	{
		do
		{
			moreblocks = SD_IDE64_ReadFileBlock(ptempchs, (struct fileblock *)pdirblock);
			if(moreblocks >= 0)
			{
				// The first entry in the first block of the directory should
				// be a label. If it isn't, then it doesn't look too promising...
				if(numberofdirblocks == 1)
				{
					if(!(pdirblock->direntry[0].attribute & (FILEATT_LABEL | FILEATT_DIR | FILEATT_LOCKED | FILEATT_CLOSED)))
						fprintf(stderr, "Warning: No label entry - [%d/%d/%d] doesn't look like a valid starting directory block...\n",ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
				}

				for (i = 0; i < NUMOFDIRENTRIES; i++)
				{
					if(pdirblock->direntry[i].attribute)
					{
						numofentries++;
					}
					else
					{
						freeentryindex = i;
						goto berlin;
					}
				}
				// IMPROVEME: Something is sick here... :-( I am almost sure this should be more simple.
				if(moreblocks)
				{
					numberofdirblocks++;

					ptempchs->cylhi = pdirblock->nextfileblock.cylhi & CLEAR2MSBITSOF8;
					ptempchs->cyllo = pdirblock->nextfileblock.cyllo;
					ptempchs->head = pdirblock->nextfileblock.head;
					ptempchs->sec = pdirblock->nextfileblock.sec;
				}
			}
			else
			{
				fprintf(stderr, "Error: SD_IDE64_DirectoryAllocEntry() - Can't read the block [%d/%d/%d]\n", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
				return -1;
			}
		} while(moreblocks > 0);

		if(numofentries % NUMOFDIRENTRIES == 0)
		{
			// IMPROVEME: should be done better instead of this dirt
			// this fills in only the necessary fields of a dummy IDE64FILE struct
			memcpy(&ide64file.lastblockchs, ptempchs, sizeof(struct chs));
			if(SD_IDE64_FileExtend(&ide64file) == 0)
			{
				memcpy(ptempchs, &ide64file.lastblockchs, sizeof(struct chs));
				freeentryindex = 0;
			}
		}
berlin:
			free(pdirblock);
	}
	else
	{
		fprintf(stderr, "Error: SD_IDE64_DirectoryAllocEntry() - Can't allocate memory for the block buffer\n");
		perror("malloc");
		return -1;
	}
	return freeentryindex;
}





/*
 * SD_IDE64_DirectoryCD2Root()
 *
 * Quickly makes the structure copy.
 * FIXME: rewrite this as macro
 */
void SD_IDE64_DirectoryCD2Root(void)
{
	memcpy(&phddinfo->cur_direntry, &phddinfo->root_direntry, sizeof(struct direntry));
	memcpy(&phddinfo->par_direntry, &phddinfo->root_direntry, sizeof(struct direntry));
	strcpy(phddinfo->currentdirpath, "/");
	return;
}





/*
 * SD_IDE64_CDPath()
 *
 * Change currentdir to the ppath given path
 * RETURNS: 0 on success, -1 on error
 */
int SD_IDE64_CDPath (char *ppath)
{
	char *ppathcomponent, *ppathbuffer; // IMPROVEME: Dynamic memory allocation based on strlen?  [512];
	struct dirblock *pdirblock;
	struct chs tempchs;
	int result = 0;
	int direntryindex;

	ppathbuffer = malloc(strlen(ppath) + 1);
	if(ppathbuffer)
	{
		// copy the given string to our nice buffer
		strcpy(ppathbuffer, ppath);
		ppath = ppathbuffer;

		// let's get rid of the leading white spaces (if any)
		while(*ppath == ' '
		|| *ppath == '\t')
			ppath++;

		// do the same to the trailing ones (if any)
		while((ppathcomponent = strrchr(ppath, ' ')) == (ppath + strlen(ppath) - 1)
		|| (ppathcomponent = strrchr(ppath, '\t')) == (ppath + strlen(ppath) - 1))
			*(ppath + strlen(ppath) -1) = '\0';

		// check if we use absolute or relative path (leading "/")
		if(*ppath == '/')
		{
			if(flags.debug)
				printf("Changing dir to: / ... ");
			SD_IDE64_DirectoryCD2Root();
			ppath++;
			if(flags.debug)
				printf("OK.\n");
		}

		//anything left of the path string?
		if((pdirblock = malloc(sizeof(struct dirblock))))
		{
			while((ppathcomponent = strsep(&ppath, "/")))
			{
				if(strlen(ppathcomponent))
				{
					if(strlen(ppathcomponent) > FILENAMELENGTH)
						fprintf(stderr, "SD_IDE64_CDPath() - Warning: path component longer than %d!\n", FILENAMELENGTH);

					if(flags.debug)
						printf("Changing dir to: %s ... ", ppathcomponent);

					if(strcmp(ppathcomponent, ".") == 0)
					{
						//IMPROVEME: Currently we do NOTHING while changing to "."! I am not sure why, but somehow I don't like it.
						if(flags.debug)
							printf("OK.\n");
						continue;
					}

					if(!strcmp(ppathcomponent, ".."))
					{
						if((memcmp(&phddinfo->par_direntry.filestartchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs))))
						{
							if((memcmp(&phddinfo->par_direntry.filestartchs, &phddinfo->root_direntry.filestartchs, sizeof(struct chs))))
							{
								memcpy(&tempchs, &phddinfo->par_direntry.filestartchs, sizeof(struct chs));
								if(SD_IDE64_BlockRead(phddinfo->pdevicefh, &tempchs, (unsigned char *)pdirblock))
								{
									direntryindex = SD_IDE64_DirectoryFindIn("..", FILEATT_DIR, &tempchs);
									if((direntryindex >= 0))
									{
										memcpy(&phddinfo->cur_direntry, &phddinfo->par_direntry, sizeof(struct direntry));
										memcpy(&phddinfo->par_direntry, &pdirblock->direntry[direntryindex], sizeof(struct direntry));
										*(strrchr(phddinfo->currentdirpath, '/')) = '\0';
										*(strrchr(phddinfo->currentdirpath, '/') + 1) = '\0';
									}
									else
									{
										if(flags.debug)
											printf("ERROR!\n");
										fprintf(stderr, "SD_IDE64_CDPath() - parent directory not found.\n");
										result = -1;
										break;
									}

								}
								else
								{
									fprintf(stderr, "SD_IDE64_CDPath() - error reading block [%d/%d/%d]\n", tempchs.cylhi * 256 + tempchs.cyllo, tempchs.head, tempchs.sec);
									result = -1;
									break;
								}

							}
							else
							{
								// quick cd to root dir
								SD_IDE64_DirectoryCD2Root();
								if(flags.debug)
									printf("OK.\n");
								continue;
							}
						}
						// IMPROVEME: cd .. on a root directory returns SUCCESS on unix, and FAILURE on IDE64
						// Not sure if I want to do anything about that.
						if(flags.debug)
							printf("OK.\n");
						continue;
					}

					direntryindex = SD_IDE64_DirectoryFindIn(ppathcomponent, FILEATT_DIR, &tempchs);
					if((direntryindex >= 0))
					{
						if(SD_IDE64_BlockRead(phddinfo->pdevicefh, &tempchs, (unsigned char *)pdirblock))
						{
							memcpy(&phddinfo->par_direntry, &phddinfo->cur_direntry, sizeof(struct direntry));
							memcpy(&phddinfo->cur_direntry, &pdirblock->direntry[direntryindex], sizeof(struct direntry));
							strcat(phddinfo->currentdirpath, ppathcomponent);
							strcat(phddinfo->currentdirpath, "/");
						}
						else
						{
							fprintf(stderr, "SD_IDE64_CDPath() - error reading block [%d/%d/%d]\n", tempchs.cylhi * 256 + tempchs.cyllo, tempchs.head, tempchs.sec);
							result = -1;
							break;
						}
					}
					else
					{
						if(flags.debug)
							printf("NOT FOUND!\n");
						fprintf(stderr, "SD_IDE64_CDPath() - path not found.\n");
						result = -1;
						break;
					}
					if(flags.debug)
						printf("OK.\n");
				}
			}
			free(pdirblock);
		}
		else
		{
			fprintf(stderr, "SD_IDE64_CDPath() - ");
			perror("malloc");
			result = -1;
		}
		free(ppathbuffer);
	}
	else
	{
		fprintf(stderr, "SD_IDE64_CDPath() - ");
		perror("malloc");
		result = -1;
	}
	return result;
}





/*
 * SD_IDE64_DirectoryMk()
 *
 * Creates new directory in the current dir
 * RETURNS: 0 on success, -1 on error.
 */
int SD_IDE64_DirectoryMk(unsigned char *pdirname)
{
	int result = 0;
	struct dirblock *pdirblock;
	struct fileblock *pfileblock;
	struct ide64node ide64node;
	unsigned char attribute, filename[FILENAMELENGTH +1];
	char *pnamepointer, *plocalnamepointer = NULL;

	pnamepointer = SD_IDE64_Filename(pdirname);
	if(pnamepointer)
	{
		if(strlen(pnamepointer))
		{
			plocalnamepointer = malloc(strlen(pnamepointer) + 1);
			if(plocalnamepointer)
				strcpy(plocalnamepointer, pnamepointer);

			if(flags.cnvexts)
			{
				attribute = SD_IDE64_Filename2FileAttByte(plocalnamepointer);
				switch(attribute)
				{
					case 0:
						break;
					case FILEATT_DIR:
						stripExtension(plocalnamepointer);
						break;
					default:
						fprintf(stderr, "SD_IDE64_DirectoryMk() - WARNING: creating dir with a valid non-dir extension!\n");
						break;
				}
			}
			// create the file
			if(!(SD_IDE64_FileCreate(plocalnamepointer, FILEATT_DIR | FILEATT_CLOSED)))
			{
				memcpy(&ide64node.dirblockchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs));
				ide64node.direntryindex = 0;
				memcpy(&ide64node.direntry.filename[0], SD_IDE64_Filename2CBMFilename(filename, plocalnamepointer), FILENAMELENGTH);
				ide64node.direntry.attribute = FILEATT_DIR;
				result = SD_IDE64_DirectoryFindEntries(&ide64node);
				if(result == 1)
				{
					result = 0;
					// alloc mem to read the first block
					pdirblock = malloc(sizeof(struct dirblock));
					if(pdirblock)
					{
						// clear the buffer... just in case.
						memset(pdirblock, 0, sizeof(struct dirblock));
						// set the block header
						pdirblock->nextfileblock.cylhi = (0 | SET2MSBITSOF8) + (sizeof(pfileblock->filedata) / 256);
						pdirblock->nextfileblock.cyllo = sizeof(pfileblock->filedata) % 256;

						// insert the obligatory dot "." ...
						pdirblock->direntry[0].attribute = FILEATT_DIR | FILEATT_LABEL | FILEATT_LOCKED | FILEATT_CLOSED;
						memset(&pdirblock->direntry[0].filename, 0xa0, sizeof(pdirblock->direntry[0].filename));
						pdirblock->direntry[0].filename[0] = '.';
						pdirblock->direntry[0].filesize[0] = 0; // this is superfluous by now as the buffer is cleared anyway
						pdirblock->direntry[0].filesize[1] = 0; // but who knows what will happen in the future...
						pdirblock->direntry[0].filesize[2] = 0; // heh?
						pdirblock->direntry[0].timestamp_sec = ide64node.direntry.timestamp_sec;
						pdirblock->direntry[0].timestamp_min = ide64node.direntry.timestamp_min;
						pdirblock->direntry[0].timestamp_hour = ide64node.direntry.timestamp_hour;
						pdirblock->direntry[0].timestamp_day = ide64node.direntry.timestamp_day;
						pdirblock->direntry[0].timestamp_month = ide64node.direntry.timestamp_month;
						pdirblock->direntry[0].timestamp_year = ide64node.direntry.timestamp_year;
						memcpy(&pdirblock->direntry[0].filestartchs, &ide64node.direntry.filestartchs, sizeof(struct chs));
						// ... and dot-dot ".." entries
						memcpy(&pdirblock->direntry[1], &phddinfo->cur_direntry, sizeof(struct direntry));
						pdirblock->direntry[1].attribute = FILEATT_DIR | FILEATT_LOCKED | FILEATT_CLOSED;
						memset(&pdirblock->direntry[1].filename, 0xa0, sizeof(pdirblock->direntry[1].filename));
						pdirblock->direntry[1].filename[0] = '.';
						pdirblock->direntry[1].filename[1] = '.';

						// write block back
						if(!(SD_IDE64_BlockWrite(phddinfo->pdevicefh, &ide64node.direntry.filestartchs, (unsigned char *)pdirblock)))
						{
							fprintf(stderr, "SD_IDE64_DirectoryMk() - error writing the dirfile block back!");
							result = -1;
						}
						free(pdirblock);
					}
					else
					{
						fprintf(stderr, "SD_IDE64_DirectoryMk() - ");
						perror("malloc");
						result = -1;
					}
				}
				else
				{
					fprintf(stderr, "SD_IDE64_DirectoryMk() - error: directory entry created but not found?!\n");
					fprintf(stderr, "SD_IDE64_DirectoryMk() - the created entry will probably be corrupted!\n");
					result = -1; // Superfluous for now. Just in case I change something later and forget
				}
			}
			else
			{
				fprintf(stderr, "SD_IDE64_DirectoryMk() - error creating directory!\n");
				result = -1;
			}
		}
		else
		{
			fprintf(stderr, "SD_IDE64_DirectoryMk() - valid filename required.\n");
			result = -1;
		}
	}
	else
	{
		fprintf(stderr, "SD_IDE64_DirectoryMk() - valid path/filename required.\n");
		result = -1;
	}
	SD_IDE64_DirectoryCD2Root();	// in case more dirs to be created, or whatever

	if(plocalnamepointer)			// IMPROVEME: Should be better structured
		free(plocalnamepointer);	// This is all about this problems with filename charsets...

	return result;
}






/*
 * SD_IDE64_FSRTraverse()
 *
 * Yeah, I couldn't stand the overcomplexity of the nonrecursive version
 * so here is the overcomplex :-D but _recursive_ version of the FSTraverse()
 * Seriously: I find this one much more easy to understand, even if probably slower
 * so I think I will sacrifice any speed gains that would be possible with the
 * nonrecursive approach. I will decide which one to use later on.
 */
int SD_IDE64_FSRTraverse(struct chs *pdirchs, struct chs *plocalpreviousdirchs, int action)
{
	int debug = 0;
	struct dirblock *pdirblock;
	struct chs tempchs, tempstartchs, *ptempchs = &tempchs;
	int numberofentriesscanned=0, numberofsubentriesscanned, moreblocks;
	int i;

	// Let's take the current dir unless another one is given
	if(pdirchs)
	{
		memcpy(ptempchs, pdirchs, sizeof(struct chs));
		memcpy(&tempstartchs, pdirchs, sizeof(struct chs));
	}
	else
	{
		memcpy(ptempchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs));
		memcpy(&tempstartchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs));
	}


	if((pdirblock = malloc(sizeof(struct dirblock))))
	{
		do
		{
			moreblocks = SD_IDE64_ReadFileBlock(ptempchs, (struct fileblock *)pdirblock);
			if(moreblocks >= 0)
			{
				for(i = 0; i < NUMOFDIRENTRIES; i++)
				{
					switch(pdirblock->direntry[i].attribute & (FILEATT_CLOSED | FILEATT_SEQ | FILEATT_PRG | FILEATT_USR | FILEATT_REL | FILEATT_DIR))
					{
						case FILEATT_CLOSED | FILEATT_SEQ:
						case FILEATT_CLOSED | FILEATT_PRG:
						case FILEATT_CLOSED | FILEATT_USR:
						{
							//	processSeqFile()
							switch(action) //FIXME: Check for error conditions in the functions below!
							{
								case TRAVERSE_DISPLAY:
									SD_IDE64_DisplayDirEntry(&pdirblock->direntry[i]);
									break;
								case TRAVERSE_BACKUP:
									SD_IDE64_FileCopy(&pdirblock->direntry[i]);
									break;
								case TRAVERSE_VALIDATE:
									SD_IDE64_FileValidate(&pdirblock->direntry[i]);
									break;
								default:
									fprintf(stderr, "SD_IDE64_FSRTraverse(SEQ) - Invalid ACTION specified!\n");
									break;
							}
							numberofentriesscanned++;
							break;
						}
						case FILEATT_CLOSED | FILEATT_REL:
						{
							//	processRelFile()
							switch(action) //FIXME: Check for error conditions in the functions below!
							{
								case TRAVERSE_DISPLAY:
									SD_IDE64_DisplayDirEntry(&pdirblock->direntry[i]);
									break;
								case TRAVERSE_BACKUP:
									SD_IDE64_FileCopy(&pdirblock->direntry[i]);
									break;
								case TRAVERSE_VALIDATE:
									SD_IDE64_FileValidate(&pdirblock->direntry[i]);
									break;
								default:
									fprintf(stderr, "SD_IDE64_FSRTraverse(REL) - Invalid ACTION specified!\n");
									break;
							}
							numberofentriesscanned++;
							break;
						}

						case FILEATT_CLOSED | FILEATT_DIR:
						{
							// We don't want to recurse into the very same directory forever ('.')
							if(!memcmp(&pdirblock->direntry[i].filestartchs, &tempstartchs, sizeof(struct chs)))
							{
								if(flags.debug && debug)
								{
									SD_IDE64_DisplayDirEntry(&pdirblock->direntry[i]);
									fprintf(stdout, "NOT traversing directory\n");
								}
								break;
							}
							// Nor we want to be pushed back to the upper directory by the '..' entry
							if(plocalpreviousdirchs)
							{
								if(!memcmp(&pdirblock->direntry[i].filestartchs, plocalpreviousdirchs, sizeof(struct chs)))
								{
									if(flags.debug && debug)
									{
										SD_IDE64_DisplayDirEntry(&pdirblock->direntry[i]);
										fprintf(stdout, "NOT traversing directory\n");
									}
									break;
								}
							}

							if(flags.debug && debug)
								fprintf(stdout, "TRAVERSING directory\n");

							switch(action) //FIXME: Check for error conditions in the functions below!
							{
								case TRAVERSE_DISPLAY:
									SD_IDE64_DisplayDirEntry(&pdirblock->direntry[i]);
									break;
								case TRAVERSE_BACKUP:
								{
									if(flags.recursive)
									{
										SD_IDE64_FileCopy(&pdirblock->direntry[i]);
										numberofsubentriesscanned = SD_IDE64_FSRTraverse(&pdirblock->direntry[i].filestartchs, &phddinfo->par_direntry.filestartchs, action);
										if(numberofsubentriesscanned >=0)
										{
											numberofentriesscanned += numberofsubentriesscanned;
										}
										else
										{
											numberofentriesscanned = -1;
											goto berlin;
										}
									}
									break;
								}
								case TRAVERSE_VALIDATE:
									SD_IDE64_FileValidate(&pdirblock->direntry[i]);
									break;
								default:
									fprintf(stderr, "SD_IDE64_FSRTraverse(DIR) - Invalid ACTION specified!\n");
									break;
							}

							numberofentriesscanned++;
							break;
						}
						default:
							break;
					}
				}
				if(moreblocks)
				{
					ptempchs->cylhi = pdirblock->nextfileblock.cylhi & CLEAR2MSBITSOF8;
					ptempchs->cyllo = pdirblock->nextfileblock.cyllo;
					ptempchs->head = pdirblock->nextfileblock.head;
					ptempchs->sec = pdirblock->nextfileblock.sec;
				}
			}
			else
			{
				numberofentriesscanned = -1;
			}
		} while(moreblocks > 0);
		if(plocalpreviousdirchs)
		{
			SD_IDE64_CDPath("..");
			if(action == TRAVERSE_BACKUP)
				chdir("..");
		}
berlin:
			free(pdirblock);
	}
	else
	{
		perror("malloc");
		numberofentriesscanned = -1;
	}
	return numberofentriesscanned;
}






/*
 * SD_IDE64_FSNewRTraverse()
 *
 * Yeah, yet another traverse version...
 * This one was created because we wanted to have the recursive
 * directory listing, which would behave like the UNIX' 'ls -R' command.
 * Therefore first scanning the whole dir and then do the second pass and
 * recurse in the subdirectories if any...
 *
 * RETURNS: number of direntries scanned.
 */
int SD_IDE64_FSNewRTraverse(char *ppath, struct chs *ppreviousdirchs, int action)
{
	struct ide64node ide64node;
	struct chs localpreviousdirchs;
	unsigned char filename[FILENAMELENGTH +1];
	char *ppathbuffer, *ptemppathbuffer, *pnamepointer;
	int i, entries = 0, entriesfound = 0;
	int numberofentriesscanned = 0, numberofsubentriesscanned;

	if(ppath)
		ppathbuffer = malloc((FILENAMELENGTH + 1 + strlen(ppath) + 1));
	else
	{
		if(flags.cnvexts)
			ppathbuffer = malloc((strlen("*.*") + 1));
		else
			ppathbuffer = malloc((strlen("*") + 1));
	}

	if(ppathbuffer)
	{
		if(ppath)
			ptemppathbuffer = malloc((FILENAMELENGTH + 1 + strlen(ppath) + 1));
		else
		{
			if(flags.cnvexts)
				ptemppathbuffer = malloc((strlen("*.*") + 1));
			else
				ptemppathbuffer = malloc((strlen("*") + 1));
		}

		if(ptemppathbuffer)
		{
			if(ppath)
				strcpy(ppathbuffer, ppath);
			else
			{
				if(flags.cnvexts)
					strcpy(ppathbuffer, "*.*");
				else
					strcpy(ppathbuffer, "*");
			}

			if(ppreviousdirchs)
				memcpy(&localpreviousdirchs, ppreviousdirchs, sizeof(struct chs));

			pnamepointer = SD_IDE64_Filename(ppathbuffer);
			if(pnamepointer)
			{
				if(strlen(pnamepointer) == 0)
				{
					if(flags.cnvexts)
						strcpy(ppathbuffer, "*.*");
					else
						strcpy(ppathbuffer, "*");
					pnamepointer = ppathbuffer;
				}
				ide64node.direntryindex = 0;
				memcpy(&ide64node.dirblockchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs));

				do
				{
					ide64node.direntry.attribute = SD_IDE64_Filename2FileAttByte(pnamepointer);
					strcpy(ptemppathbuffer, pnamepointer);

					if(flags.cnvexts && ide64node.direntry.attribute)
						stripExtension(ptemppathbuffer);
					else
						ide64node.direntry.attribute = FILEATT_ANY;

					memcpy(&ide64node.direntry.filename[0], SD_IDE64_Filename2CBMFilename(filename, ptemppathbuffer), FILENAMELENGTH);

					entriesfound = SD_IDE64_DirectoryFindEntries(&ide64node);

					if(entriesfound > 0)
					{
						switch(ide64node.direntry.attribute & (FILEATT_CLOSED | FILEATT_SEQ | FILEATT_PRG | FILEATT_USR | FILEATT_REL | FILEATT_DIR))
						{
							case FILEATT_CLOSED | FILEATT_SEQ:
							case FILEATT_CLOSED | FILEATT_PRG:
							case FILEATT_CLOSED | FILEATT_USR:
							case FILEATT_CLOSED | FILEATT_REL:
							{
								switch(action) // FIXME: Check for error conditions in the functions below!
								{
									case TRAVERSE_DISPLAY:
										SD_IDE64_DisplayDirEntry(&ide64node.direntry);
										break;
									case TRAVERSE_BACKUP:
										SD_IDE64_FileCopy(&ide64node.direntry);
										break;
									case TRAVERSE_VALIDATE:
										SD_IDE64_FileValidate(&ide64node.direntry);
										break;
									default:
										fprintf(stderr, "SD_IDE64_FSNewRTraverse(SEQ) - Invalid ACTION specified!\n");
										break;
								}
								numberofentriesscanned++;
								break;
							}
							case FILEATT_CLOSED | FILEATT_DIR:
							{
								// We don't want to recurse into the very same directory forever ('.')
								if(!memcmp(&ide64node.direntry.filestartchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs)))
								{
									if((action == TRAVERSE_DISPLAY) && flags.all)
									{
										SD_IDE64_DisplayDirEntry(&ide64node.direntry);
										numberofentriesscanned++;
									}
									break;
								}

								// Nor we want to be pushed back to the upper directory by the '..' entry
								if(!memcmp(&ide64node.direntry.filestartchs, &localpreviousdirchs, sizeof(struct chs)))
								{
									if((action == TRAVERSE_DISPLAY) && flags.all)
									{
										SD_IDE64_DisplayDirEntry(&ide64node.direntry);
										numberofentriesscanned++;
									}
									break;
								}

								switch(action) //FIXME: Check for error conditions in the functions below!
								{
									case TRAVERSE_DISPLAY:
										SD_IDE64_DisplayDirEntry(&ide64node.direntry);
										break;
									case TRAVERSE_BACKUP:
										SD_IDE64_FileCopy(&ide64node.direntry);
										break;
									case TRAVERSE_VALIDATE:
										SD_IDE64_FileValidate(&ide64node.direntry);
										break;
									default:
										fprintf(stderr, "SD_IDE64_FSNewRTraverse(DIR) - Invalid ACTION specified!\n");
										break;
								}
								numberofentriesscanned++;
								break;
							}
							default:
								break;
						}
						entries++;
						ide64node.direntryindex++;
					}
					else
					{
						if(entriesfound < 0)
							fprintf(stderr, "SD_IDE64_FSNewRTraverse() - error searching for entries.\n");
					}
				} while(entriesfound > 0);

				if(flags.recursive)
				{
					ide64node.direntryindex = 0;
					memcpy(&ide64node.dirblockchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs));
					ide64node.direntry.attribute = FILEATT_DIR;

					do
					{
						memcpy(&ide64node.direntry.filename[0], SD_IDE64_Filename2CBMFilename(filename, pnamepointer), FILENAMELENGTH);

						entriesfound = SD_IDE64_DirectoryFindEntries(&ide64node);

						if(entriesfound > 0)
						{
							switch(ide64node.direntry.attribute & (FILEATT_CLOSED | FILEATT_SEQ | FILEATT_PRG | FILEATT_USR | FILEATT_REL | FILEATT_DIR))
							{
								case FILEATT_CLOSED | FILEATT_SEQ:
								case FILEATT_CLOSED | FILEATT_PRG:
								case FILEATT_CLOSED | FILEATT_USR:
								case FILEATT_CLOSED | FILEATT_REL:
								{
									fprintf(stderr, "SD_IDE64_FSNewRTraverse(recurse) - Shouldn't find a regular file here!!!\n");
									break;
								}
								case FILEATT_CLOSED | FILEATT_DIR:
								{
									// We don't want to recurse into the very same directory forever ('.')
									if(!memcmp(&ide64node.direntry.filestartchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs)))
										break;

									// Nor we want to be pushed back to the upper directory by the '..' entry
									if(!memcmp(&ide64node.direntry.filestartchs, &localpreviousdirchs, sizeof(struct chs)))
										break;

									// OK. Let's get down on it...
									strcpy(ppathbuffer + FILENAMELENGTH + 2, pnamepointer);
									for(i = 0; i < FILENAMELENGTH; i++)
									{
										if(ide64node.direntry.filename[i] == 0xa0)
											break;
										else
											*(ppathbuffer + i) = SD_IDE64_PETASCII2ASCII(ide64node.direntry.filename[i]);
									}
									*(ppathbuffer + i) = '\0';
									strcat(ppathbuffer, "/");
									printf("\n%s", phddinfo->currentdirpath);
									fputnchars(stdout, strlen(ppathbuffer) - 1, ppathbuffer);
									printf(":\n");
									strcat(ppathbuffer, ppathbuffer + FILENAMELENGTH + 2);

									numberofsubentriesscanned = SD_IDE64_FSNewRTraverse(ppathbuffer, &phddinfo->cur_direntry.filestartchs, action);

									if(numberofsubentriesscanned >=0)
									{
										numberofentriesscanned += numberofsubentriesscanned;
										pnamepointer = filenameOnly(ppathbuffer);
									}
									else
									{
										numberofentriesscanned = -1;
										goto berlin;
									}
									break;
								}
								default:
									break;
							}
							entries++;
							ide64node.direntryindex++;
						}
						else
						{
							if(entriesfound < 0)
								fprintf(stderr, "SD_IDE64_FSNewRTraverse() - error searching for entries.\n");
						}
					} while(entriesfound > 0);
				}
			}
			else
			{
				fprintf(stderr, "SD_IDE64_FSNewRTraverse() - valid path/filename expected.\n");
				numberofentriesscanned = -1;
			}
berlin:			
			free(ptemppathbuffer);
		}
		else
		{
			fprintf(stderr, "SD_IDE64_FSNewRTraverse() - ");
			perror("malloc");
			numberofentriesscanned = -1;
		}		
			free(ppathbuffer);
	}
	else
	{
		fprintf(stderr, "SD_IDE64_FSNewRTraverse() - ");
		perror("malloc");
		numberofentriesscanned = -1;
	}

	// FIXME: Well, I know it is not the best place for these...
	if(strlen(phddinfo->currentdirpath) > 1)
		SD_IDE64_CDPath("..");
	if(action == TRAVERSE_BACKUP)
		chdir("..");

	return numberofentriesscanned;
}





/*
 * SD_IDE64_DirectoryScan() - to be replaced by something more clever soon
 *
 * Scans the directory starting at given chs
 * RETURNS: integer number of entries scanned (and displayed)
 */
int SD_IDE64_DirectoryScan(struct chs *pdirstartchs, int display)
{
	struct dirblock *pdirblock;
	struct chs tempchs, *ptempchs = &tempchs;
	int i;
	int numberofdirblocks = 1, moreblocks, numberofvalidentries = 0;

	if(pdirstartchs)
		memcpy(ptempchs, pdirstartchs, sizeof(struct chs));
	else
		memcpy(ptempchs, &phddinfo->cur_direntry.filestartchs, sizeof(struct chs));

	pdirblock = malloc(sizeof(struct dirblock));

	if(pdirblock)
	{
		do
		{
			moreblocks = SD_IDE64_ReadFileBlock(ptempchs, (struct fileblock *)pdirblock);
			if(moreblocks >= 0)
			{
				// The first entry in the first block of the directory should
				// be a label. If it isn't, then it doesn't look too promising...
				if(numberofdirblocks == 1)
				{
					if(!((pdirblock->direntry[0].attribute & (FILEATT_LABEL | FILEATT_DIR | FILEATT_LOCKED | FILEATT_CLOSED)) == (FILEATT_LABEL | FILEATT_DIR | FILEATT_LOCKED | FILEATT_CLOSED)))
						fprintf(stderr, "Warning: No label entry - [%d/%d/%d] doesn't look like a valid starting directory block...\n",ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
				}

				for (i = 0; i < NUMOFDIRENTRIES; i++)
				{
					if(pdirblock->direntry[i].attribute)
					{
						if(display)
							numberofvalidentries += SD_IDE64_DisplayDirEntry(&pdirblock->direntry[i]);
						else
							numberofvalidentries++;
					}
				}
				numberofdirblocks++;
				ptempchs->cylhi = pdirblock->nextfileblock.cylhi & CLEAR2MSBITSOF8;
				ptempchs->cyllo = pdirblock->nextfileblock.cyllo;
				ptempchs->head = pdirblock->nextfileblock.head;
				ptempchs->sec = pdirblock->nextfileblock.sec;
			}
			else
			{
				fprintf(stderr, "Error: SD_IDE64_DirectoryScan() - Can't read the block [%d/%d/%d]\n", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
				return -1;
			}
		} while(moreblocks > 0);
		free(pdirblock);
	}
	else
	{
		fprintf(stderr, "Error: SD_IDE64_DirectoryScan() - Can't allocate memory for the block buffer\n");
		perror("malloc");
		return -1;
	}
	return numberofvalidentries;
}





