/*
 * SD_IDE64_File.c
 *
 * Will hold the file related functions
 */

#include "SD_IDE64.h"

int SD_IDE64_FileRead(FILE *pdestfile, struct ide64node *pide64node)
{
	struct fileblock *pfileblock;
	struct chs tempchs, *ptempchs = &tempchs;
	int bytestocopy, moreblocks, result = 0;

	memcpy(ptempchs, &pide64node->direntry.filestartchs, sizeof(struct chs));

	pfileblock = malloc(sizeof(struct fileblock));
	if(pfileblock)
	{
		do
		{
			// Read the block and check if there is more to be read (moreblocks)
			moreblocks = SD_IDE64_ReadFileBlock(ptempchs, pfileblock);
			if(moreblocks >= 0)
			{
				if(moreblocks)
				{
					ptempchs->cylhi = pfileblock->nextfileblock.cylhi & CLEAR2MSBITSOF8;
					ptempchs->cyllo = pfileblock->nextfileblock.cyllo;
					ptempchs->head = pfileblock->nextfileblock.head;
					ptempchs->sec = pfileblock->nextfileblock.sec;
					// we have to copy the whole rest of the block
					bytestocopy = sizeof(pfileblock->filedata);
				}
				else
				{
					bytestocopy = ((((struct lastfileblock *)pfileblock)->pendingbytes[0] & CLEAR2MSBITSOF8) * 256 + ((struct lastfileblock *)pfileblock)->pendingbytes[1]);
				}

				// OK. We know where from, where to, and how many. Let's... fwrite()!
				fwrite(pfileblock->filedata, bytestocopy, 1, pdestfile);
			}
			else
			{
				fprintf(stderr, "SD_IDE64_FileRead - Error reading block [%d/%d/%d]\n", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
				result = -1;
				break;
			}
		} while(moreblocks > 0);
		free(pfileblock);
	}
	else
	{
		fprintf(stderr, "SD_IDE64_FileRead - ");
		perror("malloc");
		result = -1;
	}
	return result;
}




/*
 * SD_IDE64_FileWrite()
 *
 * Writes the source file in to the destination file starting
 * at the given chs. Returns number of bytes written or -1 on error.
 */
int SD_IDE64_FileWrite(IDE64FILE *pide64file, FILE *psrcfile)
{
	int blockswritten, numberofbyteswritten = 0, bytesread = 0;
	struct chs tempchs, *ptempchs = &tempchs, *pstartchs = &pide64file->direntry.filestartchs;
	struct lastfileblock block, *pblock = &block;

	memcpy(ptempchs, &(pide64file->direntry.filestartchs), sizeof(struct chs));

	if(SD_IDE64_BlockRead(phddinfo->pdevicefh, pstartchs, (unsigned char *)pblock))
	{
		while(!feof(psrcfile))
		{
			bytesread = fread(pblock->filedata, 1, sizeof(pblock->filedata), psrcfile);
			if(feof(psrcfile))
				break;
			else
			{
				if(bytesread == sizeof(pblock->filedata))
				{
					if(!(SD_IDE64_FileExtend(pide64file)))
					{
						// OK, extended, update link pointers
						memcpy(pblock, &pide64file->lastblockchs, sizeof(struct chs));
						*(unsigned char *)pblock = *(unsigned char *)pblock | SET1MSBITOF8; // the MSB is supposed to be always set

						// write previous buffer
						blockswritten = SD_IDE64_BlockWrite(phddinfo->pdevicefh, ptempchs, (unsigned char *)pblock);
						if(blockswritten < 1)
						{
							fprintf(stderr, "SD_IDE64_FileWrite() - error writing last block\n");
							return -1;
						}
						else
						{
							numberofbyteswritten += bytesread;
							pide64file->filesize = numberofbyteswritten;
							memcpy(ptempchs, &pide64file->lastblockchs, sizeof(struct chs));
						}
					}
					else
					{
						// error wile extending!
						fprintf(stderr, "SD_IDE64_FileWrite() - can't extend the file.\n");
						return -1;
					}
				}
				else
				{
					fprintf(stderr, "SD_IDE64_FileWrite() - ");
					perror("fread");
					return -1;
				}
			}
		}
		// IMPROVEME: Shouldn't this be in the SD_IDE64_FileClose() ?
		pblock->pendingbytes[0] = (bytesread / 256) | SET2MSBITSOF8;
		pblock->pendingbytes[1] = bytesread % 256;
		blockswritten = SD_IDE64_BlockWrite(phddinfo->pdevicefh, ptempchs, (unsigned char *)pblock);
		if(blockswritten < 1)
		{
			fprintf(stderr, "SD_IDE64_FileWrite() - error writing last block\n");
			return -1;
		}
		else
		{
			numberofbyteswritten += bytesread;
			pide64file->filesize = numberofbyteswritten;
		}
	}
	else
	{
		fprintf(stderr, "SD_IDE64_FileWrite - error reading starting fileblock: [%d/%d/%d]\n", pstartchs->cylhi * 256 + pstartchs->cyllo, pstartchs->head, pstartchs->sec);
		return -1;
	}
	return numberofbyteswritten;
}





/*
 * SD_IDE64_FileOpen()
 *
 * Performs a quasi-open to the file given by the name/attribute pair
 * RETURNS: IDE64FILE of the "open" or newly created file
 */
IDE64FILE * SD_IDE64_FileOpen(unsigned char *pfilename, unsigned char fileatt, const char *pmode)
{
	struct chs tempchs, *ptempchs = &tempchs;
	struct dirblock dirblock;
	static IDE64FILE *pide64file;
	static struct fileblock *pfileblock;
	int moreblocks, j, direntryindex;

	if(!(fileatt & FILEATT_DIR))
	{
		direntryindex = SD_IDE64_DirectoryFindIn(pfilename, fileatt, ptempchs);

		if(direntryindex >= 0 && direntryindex < NUMOFDIRENTRIES)
		{
			pide64file = malloc(sizeof(IDE64FILE));
			if(pide64file)
			{
				//clear the structure
				memset(pide64file, 0, sizeof(IDE64FILE));
				//Read the proper dir block
				if(SD_IDE64_BlockRead(phddinfo->pdevicefh, ptempchs, (unsigned char *)&dirblock))
				{
					//Fill the structure in
					memcpy(&(pide64file->direntrychs), ptempchs, sizeof(struct chs));
					pide64file->direntryindex = direntryindex;
					memcpy(&(pide64file->direntry), &dirblock.direntry[direntryindex], sizeof(struct direntry));
					for (j=0; j<3; j++)
						pide64file->filesize += dirblock.direntry[direntryindex].filesize[j] * power(256, j);
					pide64file->seekpointer = 0;
					pide64file->flags = 0; // FIXME: proper flags should be used here

					pfileblock = malloc(sizeof(struct fileblock));
					if(pfileblock)
					{
						// IMPROVEME: With this we have to read the whole file to reach its lastblock.
						memcpy(ptempchs, &pide64file->direntry.filestartchs, sizeof(struct chs));
						moreblocks = SD_IDE64_ReadFileBlock(ptempchs, pfileblock);
						if(moreblocks)
						{
							if(moreblocks > 0)
							{
								do
								{
									ptempchs->cylhi = pfileblock->nextfileblock.cylhi & CLEAR2MSBITSOF8;
									ptempchs->cyllo = pfileblock->nextfileblock.cyllo;
									ptempchs->head = pfileblock->nextfileblock.head;
									ptempchs->sec = pfileblock->nextfileblock.sec;
									moreblocks = SD_IDE64_ReadFileBlock(ptempchs, pfileblock);
									if(!(strcmp(pmode, "w")))
									{
										if(SD_IDE64_SetBAMBit(&tempchs, BLOCK_FREE))
										{
											fprintf(stderr, "SD_IDE64_FileOpen() - Error deallocating block: [%d/%d/%d]\n", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
											return NULL;
										}
									}
								} while(moreblocks > 0);
							}
							else
							{
								fprintf(stderr, "SD_IDE64_FileOpen() - error READING dirblock:\t[%d/%d/%d]... ", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
								perror(NULL);
								free(pfileblock);
								free(pide64file);
								errno = EIO;
								return NULL;
							}
						}
						if(!(strcmp(pmode, "w")))
							memcpy(&pide64file->lastblockchs, &pide64file->direntry.filestartchs, sizeof(struct chs));
						else
							memcpy(&pide64file->lastblockchs, ptempchs, sizeof(struct chs));

						memcpy(ptempchs, &pide64file->direntry.filestartchs, sizeof(struct chs));
						if(SD_IDE64_BlockRead(phddinfo->pdevicefh, ptempchs, (unsigned char *)pfileblock))
							pide64file->pseekbuffer = pfileblock;
						else
						{
							fprintf(stderr, "SD_IDE64_FileOpen() - error READING dirblock:\t[%d/%d/%d]... ", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
							perror(NULL);
							free(pfileblock);
							free(pide64file);
							errno = EIO;
							return NULL;
						}
					}
					else
					{
						fprintf(stderr, "SD_IDE64_FileOpen() - error allocating memory for seekbuffer... ");
						perror("malloc");
						free(pide64file);
						errno = ENOMEM;
						return NULL;
					}
				}
				else
				{
					fprintf(stderr, "SD_IDE64_FileOpen() - error READING dirblock:\t[%d/%d/%d]... ", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
					perror(NULL);
					free(pide64file);
					errno = EIO;
					return NULL;
				}
			}
			else
			{
				fprintf(stderr, "SD_IDE64_FileOpen() - error allocating memory for IDE64FILE... ");
				perror("malloc");
				errno = ENOMEM;
				return NULL;
			}
		}
		else
		{
			if(direntryindex >= NUMOFDIRENTRIES)
			{
				fprintf(stderr, "SD_IDE64_FileOpen() - Weird... direntry index higher than 14?!\n");
				return NULL;
			}
			if(errno == ENOENT) // Not IO error but FILE NOT FOUND.
			{
				if(!(SD_IDE64_FileCreate(pfilename, fileatt | FILEATT_CLOSED)))
				{
					pide64file = SD_IDE64_FileOpen(pfilename, fileatt, pmode);
				}
				else
				{
					fprintf(stderr, "SD_IDE64_FileOpen() - error creating: %s\n", pfilename);
					return NULL;
				}
			}
			else
			{
				return NULL;
			}
		}
	}
	else
	{
		fprintf(stderr, "SD_IDE64_FileOpen() - Can't open directory files here!\n");
		return NULL;
	}
	
	return pide64file;
}




/*
 * SD_IDE64_FileClose()
 *
 * Performs the actions necessary to "close" the file given
 * by a pointer to IDE64FILE instance.
 * RETURNS: 0 on success, -1 on errors
 */
int SD_IDE64_FileClose(IDE64FILE *pide64file)
{
	int i;
	int result = 0;
	struct dirblock dirblock;
	if(SD_IDE64_BlockRead(phddinfo->pdevicefh, &(pide64file->direntrychs), (unsigned char *)&dirblock))
	{
		if(pide64file->filesize < power(256, 3)) // 24 bits available for filelength only!
		{
			// Set the final filesize
			for(i = 2; i >= 0; i--)
				pide64file->direntry.filesize[i] = pide64file->filesize % power(256, i + 1) / power(256, i);
			// Set the FILEATT_CLOSED flag
			pide64file->direntry.attribute = pide64file->direntry.attribute | FILEATT_CLOSED;
			// copy the direntry structure into the directory block
			memcpy(&dirblock.direntry[pide64file->direntryindex], &pide64file->direntry, sizeof(struct direntry));
		}
		else
		{
			fprintf(stderr, "SD_IDE64_FileClose - File is %ld long. Too much for me now!\n", pide64file->filesize);
			errno = EFBIG;
			return EOF;
		}

		free(pide64file->pseekbuffer);

		if(!flags.readonly)
		{
			if(SD_IDE64_BlockWrite(phddinfo->pdevicefh, &(pide64file->direntrychs), (unsigned char *)&dirblock))
			{
				free(pide64file);
				result = 0;
			}
			else
			{
				fprintf(stderr, "SD_IDE64_FileClose - error WRITING the dirblock\n");
				errno = EIO;
				result = -1;
			}
		}
		else
		{
			free(pide64file);
			result = 0;
		}
	}
	else
	{
		fprintf(stderr, "SD_IDE64_FileClose - error READING the dirblock\n");
		errno = EIO;
		result = -1;
	}
	return result;
}





/*
 * SD_IDE64_FileCreate()
 *
 * Creates the file with given name/attribute pair unless it exists.
 * RETURNS: 0 on success, -1 on errors
 */
int SD_IDE64_FileCreate(unsigned char *pfilename, unsigned char fileatt)
{
	unsigned char initialclosingsequence[] = {0xc0, 0x00, 0x00, 0x00};
	struct chs tempchs, filestartchs, *ptempchs = &tempchs;
	struct dirblock dirblock, *pdirblock = &dirblock; // IMPROVEME: dynamic malloc?
	struct lastfileblock *plastfileblock;
	struct direntry *pdirentry;
	int i, freeentryindex;
	int result = 0;

	struct tm *ptime;
	time_t mytime;

	if(strlen(pfilename) > FILENAMELENGTH) // FIXME: what about the extensions?!
	{
		fprintf(stderr, "SD_IDE64_FileCreate() - WARNING: filename [%s] longer than %d! Truncating.\n", pfilename, FILENAMELENGTH);
		*(pfilename + FILENAMELENGTH) = '\0';
	}
	
	// check if the filename does(n't) exist in the current dir
	freeentryindex = SD_IDE64_DirectoryFindIn(pfilename, fileatt, &tempchs); // temporary abuse of the freeentryindex ;-()
	if(freeentryindex < 0)
	{
		freeentryindex = SD_IDE64_DirectoryAllocEntry(ptempchs);
		if(freeentryindex >= 0 && freeentryindex < NUMOFDIRENTRIES)
		{
			// read the dirblock
			if(SD_IDE64_ReadFileBlock(ptempchs, (struct fileblock *)pdirblock) >= 0)
			{
				pdirentry = &dirblock.direntry[freeentryindex];

				mytime = time(NULL);
				ptime = localtime(&mytime);

				pdirentry->attribute = fileatt;
				SD_IDE64_LogicalBlock2CHS(SD_IDE64_FindFreeBlock(), &filestartchs); // FIXME: check for error conditions!

				// Set the date of file creation
				pdirentry->timestamp_day = decimal2BCD(ptime->tm_mday);
				pdirentry->timestamp_month = decimal2BCD(ptime->tm_mon + 1); //tm_mon starts at 0, while IDE64 starts month nos at 1
				if(ptime->tm_year > 100)	// Yeah, this IS sick !
					pdirentry->timestamp_year = decimal2BCD(ptime->tm_year - 100);
				else
					pdirentry->timestamp_year = decimal2BCD(ptime->tm_year);

				// Set the time of file creation
				pdirentry->timestamp_sec = decimal2BCD(ptime->tm_sec);
				pdirentry->timestamp_min = decimal2BCD(ptime->tm_min);
				pdirentry->timestamp_hour = decimal2BCD(ptime->tm_hour);

				// Translate and set the filename - $a0 delimited
				memset(pdirentry->filename, 0xa0, sizeof(pdirentry->filename));
				for(i = 0; i < strlen(pfilename); i++)
					pdirentry->filename[i] = SD_IDE64_ASCII2PETASCII(*(pfilename + i));
				//					pdirentry->filename[i] = topetascii[*(pfilename + i)];

				// Set the starting block of the file
				memcpy(&(pdirentry->filestartchs), &filestartchs, sizeof(struct chs));

				// Set the initial file size (to 0 bytes)
				memset(pdirentry->filesize, 0, sizeof(pdirentry->filesize));
				
				//set the BAM bit for the first file block
				if(SD_IDE64_SetBAMBit(&pdirentry->filestartchs, BLOCK_USED))
				{
					//initialise the first file block
					plastfileblock = malloc(sizeof(struct lastfileblock));
					if(plastfileblock)
					{
						// Initialising the block with 0s
						memset(plastfileblock, 0, sizeof(struct lastfileblock));
						// Updating the nextfileblock pointer to reflect the LAST block status
						memcpy(plastfileblock, initialclosingsequence, sizeof(struct chs));
						if(SD_IDE64_BlockWrite(phddinfo->pdevicefh, &filestartchs, (unsigned char *)plastfileblock))
						{
							//write dirblock back
							if(!SD_IDE64_BlockWrite(phddinfo->pdevicefh, ptempchs, (unsigned char *)pdirblock) > 0)
							{
								fprintf(stderr, "SD_IDE64_FileCreate() - Error writing back the dirblock:\t[%d/%d/%d]\n", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
								result = -1;
							}
						}
						else
						{
							fprintf(stderr, "SD_IDE64_FileCreate() - Error the first file block:\t[%d/%d/%d]\n", filestartchs.cylhi * 256 + filestartchs.cyllo, filestartchs.head, filestartchs.sec);
							result = -1;
						}

						free(plastfileblock);
					}
					else
					{
						fprintf(stderr, "SD_IDE64_FileCreate() - ");
						perror("malloc");
						result = -1;
					}
				}
				else
				{
					fprintf(stderr, "SD_IDE64_FileCreate() - Error setting the BAM bit for block:\t[%d/%d/%d]\n", pdirentry->filestartchs.cylhi * 256 + pdirentry->filestartchs.cyllo, pdirentry->filestartchs.head, pdirentry->filestartchs.sec);
					result = -1;
				}

			}
			else
			{
				fprintf(stderr, "SD_IDE64_FileCreate() - Error reading the dirblock:\t[%d/%d/%d]\n", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
				result = -1;
			}
		}
		else
		{
			if(freeentryindex < 0)
				fprintf(stderr, "SD_IDE64_FileCreate() - No direntry allocated. Out of free blocks?\n");
			else
				fprintf(stderr, "SD_IDE64_FileCreate() - Weird... direntry index higher that 14?!\n");
			result = -1;
		}
	}
	else
	{
		fprintf(stderr, "SD_IDE64_FileCreate() - FILE EXISTS!\n");
		errno = EEXIST;
		result = -1;
	}
	return result;
}




/*
 * SD_IDE64_FileExtend()
 *
 * Takes the (currently) last fileblock chs pointer, finds next free
 * block, allocates it and updates the pointer to return the next block
 * RETURNS: 0 on success, -1 on errors
 */
int SD_IDE64_FileExtend(IDE64FILE *pide64file)
{
	struct chs tempchs, *ptempchs = &tempchs, *pfilelastchs = &(pide64file->lastblockchs);
	long int logicalblock;
	int i;
	unsigned char initialclosingsequence[] = {0xc0, 0x00, 0x00, 0x00};
	struct lastfileblock *plastfileblock;

	logicalblock = SD_IDE64_FindFreeBlock();
	if(logicalblock > 0)
	{
		if(SD_IDE64_LogicalBlock2CHS(logicalblock, ptempchs))
		{
			//set the BAM bit for this block
			if(SD_IDE64_SetBAMBit(ptempchs, BLOCK_USED))
			{
				plastfileblock = malloc(sizeof(struct lastfileblock));
				if(plastfileblock)
				{
					if(SD_IDE64_BlockRead(phddinfo->pdevicefh, ptempchs, (unsigned char *)plastfileblock))
					{
						// Initialising the block with 0s
						memset(plastfileblock, 0, sizeof(struct lastfileblock));
						// Updating the nextfileblock pointer to reflect the LAST block status
						memcpy(plastfileblock, initialclosingsequence, sizeof(struct chs));
						if(SD_IDE64_BlockWrite(phddinfo->pdevicefh, ptempchs, (unsigned char *)plastfileblock))
						{
							if(SD_IDE64_BlockRead(phddinfo->pdevicefh, pfilelastchs, (unsigned char *)plastfileblock))
							{
								// Linking the previously last block with the current LAST one
								for(i = 0; i < sizeof(struct chs); i++)
								{
									if(i == 0)
										*((unsigned char *)plastfileblock + i) = *((unsigned char *)ptempchs + i) | SET1MSBITOF8; // the MSB is supposed to be always set
									else
										*((unsigned char *)plastfileblock + i) = *((unsigned char *)ptempchs + i);
								}
								if(!SD_IDE64_BlockWrite(phddinfo->pdevicefh, pfilelastchs, (unsigned char *)plastfileblock))
								{
									fprintf(stderr, "SD_IDE64_FileExtend() - linking blocks [%d/%d/%d]: error writing the block back.\n", pfilelastchs->cylhi * 256 + pfilelastchs->cyllo, pfilelastchs->head, pfilelastchs->sec);
									return -1;
								}
							}
							else
							{
								fprintf(stderr, "SD_IDE64_FileExtend() - linking blocks [%d/%d/%d]: error reading the block.\n", pfilelastchs->cylhi * 256 + pfilelastchs->cyllo, pfilelastchs->head, pfilelastchs->sec);
								return -1;
							}
						}
						else
						{
							fprintf(stderr, "SD_IDE64_FileExtend() - making block the LAST one [%d/%d/%d]: error writing the block back.\n", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
							return -1;
						}
					}
					else
					{
						fprintf(stderr, "SD_IDE64_FileExtend() - making block the LAST one [%d/%d/%d]: error reading the block.\n", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
						return -1;
					}
					free(plastfileblock);
				}
				else
				{
					fprintf(stderr, "SD_IDE64_FileExtend() - ");
					perror("malloc");
					return -1;
				}
			}
			else
			{
				fprintf(stderr, "SD_IDE64_FileExtend() - error setting the BAM bit for [%d/%d/%d]: Invalid block number?\n", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
				return -1;
			}
		}
		else
		{
			fprintf(stderr, "SD_IDE64_FileExtend() - allocating block: %ld\t[%d/%d/%d]: Invalid block number?\n", logicalblock, ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
			return -1;
		}
	}
	else
	{
		fprintf(stderr, "SD_IDE64_FileExtend() - Can't find free block! Out of space?\n");
		return -1;
	}
	// All done? Really? Then let's update the IDE64FILE...
	memcpy(&(pide64file->lastblockchs), ptempchs, sizeof(struct chs));
	return 0;
}





/*
 * SD_IDE64_fread()
 *
 * LAME implementation of basic fread done on IDE64FILE
 * The longer I look on it, the more I believe it has to be trashed
 * and rewritten from scratch :-(
 */
size_t SD_IDE64_fread(void *ptr, size_t size, size_t nmemb, IDE64FILE *pide64file)
{
	int debug = 0;
	int lastblock = 0;
	static long int lastaccessedfileblocknumber = 0, currentbytefileblocknumber;
	short int bytesinthisblock;
	unsigned long int filebyteoffset;
	struct chs tempchs, *ptempchs = &tempchs;

	int i, j;

	if(*(unsigned char *)pide64file->pseekbuffer & FILE_LASTBLOCK)
	{
		lastblock = 1;
		bytesinthisblock = (*(unsigned char *)pide64file->pseekbuffer & CLEAR2MSBITSOF8) * 256 + *((unsigned char *)pide64file->pseekbuffer + 1);
	}
	else
		bytesinthisblock = sizeof(pide64file->pseekbuffer->filedata);

	for(i=0; i<nmemb; i++)
	{
		for(j=0; j<size; j++)
		{
			currentbytefileblocknumber = pide64file->seekpointer / sizeof(pide64file->pseekbuffer->filedata);
			filebyteoffset = pide64file->seekpointer % sizeof(pide64file->pseekbuffer->filedata);

			if((currentbytefileblocknumber == (lastaccessedfileblocknumber + 1)))
			{
				if(!(lastblock))
				{
					ptempchs->cylhi = pide64file->pseekbuffer->nextfileblock.cylhi & CLEAR2MSBITSOF8;
					ptempchs->cyllo = pide64file->pseekbuffer->nextfileblock.cyllo;
					ptempchs->head = pide64file->pseekbuffer->nextfileblock.head;
					ptempchs->sec = pide64file->pseekbuffer->nextfileblock.sec;

					if(SD_IDE64_BlockRead(phddinfo->pdevicefh, ptempchs, (unsigned char *)pide64file->pseekbuffer))
					{
						lastaccessedfileblocknumber = currentbytefileblocknumber;

						if(*(unsigned char *)pide64file->pseekbuffer & FILE_LASTBLOCK)
						{
							lastblock = 1;
							bytesinthisblock = (*(unsigned char *)pide64file->pseekbuffer & CLEAR2MSBITSOF8) * 256 + *((unsigned char *)pide64file->pseekbuffer + 1);
						}
						else
							bytesinthisblock = sizeof(pide64file->pseekbuffer->filedata);
					}
					else
					{
						fprintf(stderr, "SD_IDE64_fread() - Error reading file block.\n");
						errno = EIO;
						goto berlin;
					}
				}
				else
					bytesinthisblock = 0;
			}

			if(filebyteoffset < bytesinthisblock)
			{
				*((unsigned char *)ptr + (i * size) + j) = pide64file->pseekbuffer->filedata[filebyteoffset];
				pide64file->seekpointer++;
			}
			else
			{
				if(flags.debug && debug)
					fprintf(stdout, "SD_IDE64_fread() - EOF reached at:\t%ld\n", pide64file->seekpointer);
				lastaccessedfileblocknumber = 0;
				goto berlin;	//EOF reached!
			}
		}
	}
berlin:
		return i;
}





int SD_IDE64_FileCopy(struct direntry *pdirentry)
{
	unsigned char filenamebuffer[FILENAMELENGTH + 1 + 3 + 1 + 1]; // FILENAME + "." + EXTENSION + LOCKED? + \0
	IDE64FILE *pide64file;
	FILE *pfile;
	unsigned char tempbyte;
	int i;

	char *nullbuff = NULL;

	if(pdirentry->attribute & FILEATT_CLOSED)
	{
		//FIXME: this is a temporary CRAP! Needed because the SD_IDE64_DirectoryFindIn() requires ASCII instead of PETASCII...
		for(i = 0; i < FILENAMELENGTH; i++)
		{
			if(pdirentry->filename[i] == 0xa0)
				break;
			else
				filenamebuffer[i] = SD_IDE64_PETASCII2ASCII(pdirentry->filename[i]);
		}
		filenamebuffer[i] = '\0';
		//until here... hope it won't last forever!

		if(pdirentry->attribute & FILEATT_DIR)
		{
			if(flags.verbose)
				printf("Changing IDE64 directory to:\t%s... ", filenamebuffer);
			if(SD_IDE64_CDPath(filenamebuffer) == 0)
			{
				if(flags.verbose)
					printf("OK.\n");
				SD_IDE64_CBMFilename2Filename(filenamebuffer, pdirentry->filename, pdirentry->attribute);
				if(flags.verbose)
					printf("Creating directory:\t%s... ", filenamebuffer);
				if(!mkdir(filenamebuffer, 0755))
				{
					if(flags.verbose)
						printf("OK. Getting there... ");
					if(chdir(filenamebuffer))
					{
						if(flags.verbose)
							printf("BAD LUCK!\n");
						fprintf(stderr, "SD_IDE64_FileCopy() - error changing dir to:\t%s\n", filenamebuffer);
						return -1;
					}
					else
					{
						if(flags.verbose)
							printf("OK.\n");
					}
				}
				else
				{
					if(flags.verbose)
						printf("BAD LUCK!\n");
					fprintf(stderr, "SD_IDE64_FileCopy() - ");
					perror("mkdir");
					return -1;
				}
				if(flags.debug)
				{
					printf("Current dir:\t%s\n", getcwd(nullbuff, 0));
					free(nullbuff);
				}
			}
			else
			{
				if(flags.verbose)
					printf("BAD LUCK!\n");
				fprintf(stderr, "SD_IDE64_FileCopy() - error changing IDE64 directory");
				return -1;
			}
			return 0;
		}

		if((pide64file = SD_IDE64_FileOpen(filenamebuffer, pdirentry->attribute, "r")))
		{
			SD_IDE64_CBMFilename2Filename(filenamebuffer, pdirentry->filename, pdirentry->attribute);
			if((pfile = fopen(filenamebuffer, "w")))
			{
				if(flags.verbose)
					printf("Copying file:\t%s... ", filenamebuffer);
				while(SD_IDE64_fread(&tempbyte, 1, 1, pide64file))
				{
					if(!fwrite(&tempbyte, 1, 1, pfile))
					{
						if(flags.verbose)
							printf("BAD LUCK!\n");
						fprintf(stderr, "SD_IDE64_FileCopy() - ");
						perror("fwrite");
						return -1;
					}
				}
				if(!fclose(pfile))
				{
					if(flags.verbose)
						printf("OK.\n");
				}
				else
				{
					if(flags.verbose)
						printf("BAD LUCK!\n");
					fprintf(stderr, "SD_IDE64_FileCopy() - ");
					perror("fclose");
				}
			}
			else
			{
				fprintf(stderr, "SD_IDE64_FileCopy() - ");
				perror("fopen");
				return -1;
			}
			SD_IDE64_FileClose(pide64file); //FIXME: check for errors on closing
		}
		else
		{
			fprintf(stderr, "SD_IDE64_FileCopy() - error opening IDE64 file!\n");
			return -1;
		}
	}
	else
	{
		if(flags.verbose)
		{
			printf("Warning!/n");
			SD_IDE64_DisplayDirEntry(pdirentry);
			printf("Seems to be open?!/n");
		}
		return -1;
	}
	return 0;
}







int SD_IDE64_FileValidate(struct direntry *pdirentry)
{
	return 0;
}





/*
 * SD_IDE64_FileFreeBlocks()
 *
 * Scans linked blocks starting at given chs and sets them free in the BAM
 * RETURNS: 0 on success, -1 on errors.
 */
int SD_IDE64_FileFreeBlocks(struct chs *pstartchs)
{
	int blocksfreed = 0, moreblocks, result;
	struct chs tempchs, *ptempchs = &tempchs;
	struct fileblock *pfileblock;

	memcpy(ptempchs, pstartchs, sizeof(struct chs));

	pfileblock = malloc(sizeof(struct fileblock));
	if(pfileblock)
	{
		do
		{
			// Read the block and check if there is more to be read (moreblocks)
			moreblocks = SD_IDE64_ReadFileBlock(ptempchs, pfileblock);
			if(moreblocks >= 0)
			{
				result = SD_IDE64_GetBAMBit(ptempchs);
				if(result)
				{
					result = SD_IDE64_SetBAMBit(ptempchs, BLOCK_FREE);
					if(result)
					{
						fprintf(stderr, "SD_IDE64_FileFreeBlocks - Error freeing block [%d/%d/%d]\n", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
						result = -1;
						break;
					}
					blocksfreed++;
				}
				else
				{
					fprintf(stderr, "SD_IDE64_FileFreeBlocks - ERROR: block [%d/%d/%d] already free! Corrupted file?\n", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
					result = -1;
					break;
				}
				if(moreblocks)
				{
					ptempchs->cylhi = pfileblock->nextfileblock.cylhi & CLEAR2MSBITSOF8;
					ptempchs->cyllo = pfileblock->nextfileblock.cyllo;
					ptempchs->head = pfileblock->nextfileblock.head;
					ptempchs->sec = pfileblock->nextfileblock.sec;
				}
			}
			else
			{
				fprintf(stderr, "SD_IDE64_FileFreeBlocks - Error reading block [%d/%d/%d]\n", ptempchs->cylhi * 256 + ptempchs->cyllo, ptempchs->head, ptempchs->sec);
				result = -1;
				break;
			}
		} while(moreblocks > 0);
		free(pfileblock);
	}
	else
	{
		fprintf(stderr, "SD_IDE64_FileFreeBlocks - ");
		perror("malloc");
		result = -1;
	}
	if(flags.verbose)
		printf("Blocks deallocated:\t%d\n", blocksfreed);
	return result;
}





/*
 * SD_IDE64_NodeDelete()
 *
 * Deletes the node described by the given ide64node
 * RETURNS: 0 on success, -1 on error.
 */
int SD_IDE64_NodeDelete(struct ide64node *pide64node)
{
	int result = 0;
	int direntryindex;
	struct chs tempchs;
	struct dirblock *pdirblock;

	direntryindex = pide64node->direntryindex;
	memcpy(&tempchs, &pide64node->dirblockchs, sizeof(struct chs));
	pdirblock = malloc(sizeof(struct dirblock));
	if(pdirblock)
	{
		if(SD_IDE64_BlockRead(phddinfo->pdevicefh, &tempchs, (unsigned char *)pdirblock))
		{
			// mark the direntry[direntryindex] as DEL-eted
			pdirblock->direntry[direntryindex].attribute = FILEATT_DEL;
			// write the dirblock back
			if(SD_IDE64_BlockWrite(phddinfo->pdevicefh, &tempchs, (unsigned char *)pdirblock))
			{
				// read all the file blocks and mark them as BLOCK_FREE in the BAM
				memcpy(&tempchs, &pdirblock->direntry[direntryindex].filestartchs, sizeof(struct chs));
				result = SD_IDE64_FileFreeBlocks(&tempchs);
				if(result)
				{
					fprintf(stderr, "SD_IDE64_NodeDelete() - WARNING: freeing the block chain unsuccessful!\n");
				}
			}
			else
			{
				fprintf(stderr, "SD_IDE64_NodeDelete() - error writing back dirblock [%d/%d/%d]\n", tempchs.cylhi * 256 + tempchs.cyllo, tempchs.head, tempchs.sec);
				result = -1;
			}
		}
		else
		{
			fprintf(stderr, "SD_IDE64_NodeDelete() - error reading dirblock [%d/%d/%d]\n", tempchs.cylhi * 256 + tempchs.cyllo, tempchs.head, tempchs.sec);
			result = -1;
		}
	}
	else
	{
		fprintf(stderr, "SD_IDE64_NodeDelete() - ");
		perror("malloc");
		result = -1;
	}
	return result;
}






/*
 * SD_IDE64_fopen()
 *
 * This should be somewhat less LAME implementation.
 * Takes the path and mode parameters for input.
 * RETURNS: pointer to the ide64file structure or NULL
 */
IDE64FILE * SD_IDE64_fopen(const char *ppath, const char *pmode)
{
	IDE64FILE *pide64file;
	unsigned char *pnamepointer;

	pnamepointer = SD_IDE64_Filename(ppath);
	if(pnamepointer)
	{
		if(strlen(pnamepointer) > 0)
		{
			if(strpbrk(pmode, "rwa"))
			{
				if(*pmode == 'b')
					pmode++;
				pide64file = malloc(sizeof(IDE64FILE));
				if(pide64file)
				{
					// check if exists
					if(*pmode == 'r')
					{

					}
				}
				else
				{
					fprintf(stderr, "SD_IDE64_fopen() - ");
					perror("malloc");
					//pide64file = NULL;
				}
			}
			else
			{
				fprintf(stderr, "SD_IDE64_fopen() - error! \"%s\" Invalid mode?\n", pmode);
				pide64file = NULL;
			}
		}
		else
		{
			fprintf(stderr, "SD_IDE64_fopen() - valid filename required!\n");
			pide64file = NULL;
		}
	}
	else
	{
		fprintf(stderr, "SD_IDE64_fopen() - error! \"%s\" Invalid path?\n", ppath);
		pide64file = NULL;
	}
	return pide64file;
}




