/*
 * SD_IDE64_Structures.h
 *
 * This file contains the definition of known
 * IDE64 structures.
 * It is expected to grow to cover everything what's
 * necessary for writing a working reimplementation
 * of the ide64's filesystem
 *
 */


#ifndef SD_IDE64_STRUCTURES_H
#define SD_IDE64_STRUCTURES_H


/*
 * The basic structure for sectors addressing
 */
struct chs
{
	unsigned char		cylhi;
	unsigned char		cyllo;
	unsigned char		head;
	unsigned char		sec;
};





/*
 * Raw sector structure - might get handy some day...
 */
struct hddblock
{
	unsigned char		blockbyte[512];
};



/*
 * This is the first block of the drive. If it gets corrupted - bad luck!
 */
struct rootblock
{
	unsigned char		identification[16];

	struct chs			lastdriveblock;
	struct chs			rootdirblock;
	struct chs			bam1block;
	struct chs			bam2block;

	unsigned char		reserved[0x0200 - 0x0020];
};




/*
 * We are going to use the direntries in the
 * directory herself, later on...
 */
struct direntry
{
	unsigned char		attribute;
	unsigned char		reserved0;
	unsigned char		filename[16];
	unsigned char		filesize[3];
	unsigned char		reserved1;
	unsigned char		timestamp_sec;
	unsigned char		timestamp_min;
	unsigned char		timestamp_hour;
	unsigned char		timestamp_day;
	unsigned char		timestamp_month;
	unsigned char		timestamp_year;

	struct chs			filestartchs;
};





/*
 * ... like in this block. It is a directory block and
 * has 15 direntried starting at the offset $0008.
 */
struct dirblock
{
	struct chs			nextfileblock;
	unsigned char		unused0[4];
	struct direntry		direntry[15];
	unsigned char		unused1[24];
};





/*
 * a temporary node structure. Will have to be replaced by a real inode later.
 */
struct ide64node
{
	struct chs dirblockchs;
	int direntryindex;
	struct direntry direntry;
};





/*
 * This is a sector allocated for a file
 */
struct fileblock
{
	struct chs			nextfileblock;
	unsigned char		filedata[508]; /* 512 - 4 (sizeof(fileblock header bytes) */
};


/*
 * This is the last sector allocated for a file
 */
struct lastfileblock
{
	unsigned char		pendingbytes[2]; /* don't forget setting two MSBits (7th always set, 6th declaring LAST file block) */
	unsigned char		unused0;
	unsigned char		unused1;
	unsigned char		filedata[508]; /* 512 - 4 (sizeof(fileblock header bytes) */
};


/*
 * struct hddinfo
 * the structure holding all the most important info about the drive (image)
 */
struct hddinfo
{
	int blocksize;
	FILE *pdevicefh;
	struct stat sb;
	struct group* pgroup;
	struct passwd* ppasswd;
	unsigned char identification[16];
	struct chs geometry;
	struct chs bam1start;
	struct chs bam2start;
	struct chs rootdirchs;
	struct direntry root_direntry;
	struct direntry cur_direntry;
	struct direntry par_direntry;
	unsigned long int numberoflogicalblocks;
	unsigned char currentdirpath[512];
};





#endif
