/*
 * SD_IDE64_main.c
 *
 * What could this be?
 */

#include "SD_IDE64.h"

// Global program flags
struct binaryflags flags;

// Global HDD information
struct hddinfo* phddinfo;

char *outputname, *devicename;	/* XXX (silverdr) Let's have this here for a while. */




/* describes one command */
struct command {
    char *name;					/* name of command - from commandline */
    int (*function)(int argc, char*argv[]);	/* function called with the rest of parameters */
};



/* command list - add commands here */
struct command commands[] = {
    /* real commands */
	{ "backup", &DoBackup },
	{ "mkdir", &DoMkdir },
	{ "info", &DoInfo },
    /* experimental commands */
	{ "ls", &DoLs },
	{ "rm", &DoRm },
	{ "rmdir", &DoRmDir },
	{ "read", &DoRead },
	{ "write", &DoWrite },
	/* unimplemented commands */
	{ "mv", &DoMv },
	{ "restore", &DoRestore },
	{ "mkfs", &DoMkfs },
	{ "fsck", &DoFsck },
	{ "test", &DoTest },
	{ NULL, NULL } };

/* actual program */

void show_usage(const char *progname)
{
    int i=0;

    printf("Usage: %s [options] command [command arguments]\n\n"
	   "  -p, --preserve-attributes\tenable attribute preserving\n"
	   "  -a, --all-dirs\tInclude directory entries whose names begin with a dot (.)\n"
	   "  -c, --no-convert-extensions\tdisable converting the CBM file attributes into extensions\n"
	   "  -n, --no-writes\t\tno writes to the disk/image will be done\n"
	   "  -l, --long-dir\tdirectory list in long format\n"
	   "  -R, --recursive\t\tenable recursive behavior\n"
	   "  -d, --device-name=NAME\tname for block device or IDE64 image file (required)\n"
	   "  -o, --output-name=NAME\tname for backup directory\n\n"
	   "  -h, --help\t\t\tthis help\n\nAvailable commands:\n  ", progname);
    while (commands[i].name)
	printf("%s ",commands[i++].name);
    printf("\n");
}

int main (int argc, char * argv[])
{
	int i, c;
	int remained, retval, option_index = 0;

	/* set the default values */
	flags.help = 0;
	flags.verbose = 0;
	flags.debug = 0;
	flags.error = 0;
	flags.readonly = 0;
	flags.cnvexts = 1;
	flags.preserve = 0;
	flags.recursive = 0;
	flags.longdir = 0;
	flags.all = 0;
	outputname = NULL;
	devicename = NULL;

	/* if no arguments - show usage */
	if (argc==1) {
	    show_usage(argv[0]);
	    return EXIT_SUCCESS;
	}

	while (1) {
		static struct option long_options[] = {
			/* options */
			{"verbose", no_argument, &flags.verbose, 1},
			{"no-writes", no_argument, &flags.readonly, 1},
			{"recursive", no_argument, &flags.recursive, 1},
			{"long-dir", no_argument, &flags.longdir, 1},
			{"all-dirs", no_argument, &flags.all, 1},
			{"no-convert-extensions", no_argument, &flags.cnvexts, 0},
			{"preserve-attributes", no_argument, &flags.preserve, 1},
			{"output-name", required_argument, 0, 'o'},
			{"device-name", required_argument, 0, 'd'},
			{"help", no_argument, 0, 'h'},
			{0,0,0,0} };

		c = getopt_long(argc,argv, "vnRlcpao:d:h", long_options, &option_index);

		if (c==-1) break;	/* no more options */

		switch (c) {
			case 0:
				fprintf(stderr, "getopt returned 0 - error in long_options definition\n");
				break;
			case 'v':
				flags.verbose = 1;
				break;
			case 'n':
				flags.readonly = 1;
				break;
			case 'R':
				flags.recursive = 1;
				break;
			case 'l':
				flags.longdir = 1;
				break;
			case 'a':
				flags.all = 1;
				break;
			case 'c':
				flags.cnvexts = 0;
				break;
			case 'p':
				flags.preserve = 1;
				break;
			case 'o':
				outputname = optarg;
				break;
			case 'd':
				devicename = optarg;
				break;
			case 'h':
				show_usage(argv[0]);
				return EXIT_SUCCESS;
			default:
				fprintf (stderr,"-%c: Unrecognized option\n", c);
				return EXIT_FAILURE;
			}
	}

	if (!devicename) {
	    fprintf (stderr, "No IDE64 image/device name, use -d or --device-name\n");
	    return EXIT_FAILURE;
	}

	if (optind == argc) {
	    fprintf(stderr, "No commands provided!\n");
	    return EXIT_FAILURE;
	}

	/* all commands require drive to be logged in */

	i=0;
	remained = argc-optind - 1;
	while (commands[i].name)
	{
		if (!strcmp(argv[optind], commands[i].name))
		{
			if(flags.verbose)
				printf("Logging the drive in:\n");
			if(SD_IDE64_DriveLogin(devicename) == 0)
			{
				if(flags.verbose)
					printf("Done.\n\n");

				// This is the real stuff
				retval = commands[i].function(remained, &argv[optind+1]);
				// and here it ends

				if(flags.verbose)
					printf("\nLogging the drive out:\n");

				if(SD_IDE64_DriveLogout())
				{
					fprintf(stderr, "SD_IDE64_Main() - error logging out the drive!\n");
					retval = EXIT_FAILURE;
				}
				if(flags.verbose)
					printf("READY.\n");
				return retval;
			}
			else
			{
				fprintf(stderr, "SD_IDE64_Main() - can't log the selected drive in!\n");
				return EXIT_FAILURE;
			}
		}
		i++;
	}

	fprintf(stderr, "SD_IDE64_Main() - unknown command: %s\n", argv[optind]);
	printf("Available commands:\n  ");
	i = 0;
    while (commands[i].name)
		printf("%s ",commands[i++].name);
    printf("\n");
	return EXIT_FAILURE;

}
